package assign;

import input.Protein;

import java.io.Serializable;
import java.lang.ref.SoftReference;
import java.util.Arrays;

import util.BitSet;
import util.Util;
import util.Util.Pair;

// DEBUG
import EDU.oswego.cs.dl.util.concurrent.SynchronizedInt;

public class Solution {

	private Window[] windows;
	private BitSet[] summary;			// null means it needs to be recomputed
	private BitSet fixedPos;
	private WindowInfo[] parentWindowInfo;
	//context
	private static Protein protein = Worker.protein;

	private static Solution root;
	
	// DEBUG
	private static SynchronizedInt instanceCount = new SynchronizedInt(0);

	//-----------------------------------------------------------------------------------
	//-----------------------------CONSTRUCTORS------------------------------------------
	//-----------------------------------------------------------------------------------
	//---------------------------set of windows of length 1------------------------------
	public Solution() {

		windows = new Window[protein.getLength()];
		for (int pos = 0; pos < protein.getLength(); pos++)
			windows[pos] = new Window(pos);

		fixedPos = new BitSet(0, protein.getLength());
		summary = null;
		
		instanceCount.increment();
	}

	//--------produce a subsolution by fixing a spin system at a position--------------
	public Solution(Solution solution, int spinSystemId, int pos) throws DeadEnd {
		
		parentWindowInfo = new WindowInfo[solution.getWindows().length];
		windows = new Window[solution.getWindows().length];
		for (int i = 0; i < solution.getWindows().length; i++) {
			windows[i] = new Window(solution.getWindows()[i], spinSystemId, pos);
			parentWindowInfo[i] = new WindowInfo(solution.getWindows()[i]);
		}

		fixedPos = solution.getFixedPos().copy();
		fixedPos.set(pos);
		summary = null;
		
		instanceCount.increment();
	}

	//------reproduce a subsolution by fixing several spin systems from the root-------
	private Solution(int[] spinSystemIds, int[] positions, WindowInfo[] parentWindowInfo) throws DeadEnd {

		this.parentWindowInfo = parentWindowInfo;
		fixedPos = root.getFixedPos().copy();
		windows = new Window[root.getWindows().length];
		
		for (int i = 0; i < windows.length; i++)
			windows[i] = new Window(root.getWindows()[i], spinSystemIds, positions);

		for (int i = 0; i < spinSystemIds.length; i++)
			fixedPos.set(positions[i]);
		
		summary = null;
		
		instanceCount.increment();
	}

	//-----------------------------------------------------------------------------------
	//----------------------------INITIALIZATION-----------------------------------------
	//-----------------------------------------------------------------------------------

	//-----------------create the root solution of the search tree-----------------------
	public static void initialize() {
	
		//strands of length 1 at each position
		root = new Solution();
		Util.pln("\n\nINITIALIZATION:\n" + root.toStringByPositions());
	
                Util.pln("\n\nEXTEND WINDOWS:");
		//extend windows and base case: dead end
		try {
			root.mergeWindows();
		} catch (Solution.DeadEnd e) {
			Util.pln("Dead end at the initial merge stage. No assignments can be found. The program will now stop.");
			System.exit(0);
		}
		Util.pln("\nAFTER PRECOMPUTATION:\n" + root.toStringByPositions());
	}

	//----------------------------------------------------------------------------------
	//--------------------------------MUTATORS------------------------------------------
	//----------------------------------------------------------------------------------

	//---------------------merge all windows with no checks------------------------------
	public void mergeWindows() throws DeadEnd {

		//window by window
		for (int i = 0; i < windows.length - 1; i++) {

			try {
				windows[i
					+ 1].mergeWithPrevious(
						windows[i],
						computeMinMissingAround(i, i + 1),
						computeMinScoreAround(i, i + 1));
				windows[i] = null;
			} catch (OutOfMemoryError e) {
				Util.pln("Gave up out of memory");
			} catch (MaxSizeExceeded e) {
				Util.pln("MaxSizeExceeded");
			}
		}
		windows = (Window[]) Util.stripEmpties(windows);
		
		summary = null;
	}

	//---------------------merge highly modified windows--------------------------------
	public void checkAndMergeWindows() throws DeadEnd {
		boolean[] failedToMerge = new boolean[getWindows().length];

		while (true) {

			int minNumberStrands = Integer.MAX_VALUE;
			int windowId = -1;
			int nextWindowId = -1;

			for (int i = 0; i < getWindows().length - 1; i++) {

				Window w = getWindows()[i];
				if (failedToMerge[i])
					continue;

				int numberStrands = -1;
				int j = i + 1;
				for (; j < getWindows().length; j++)
					if (getWindows()[j] != null) {
						numberStrands = getWindows()[j].getStrands().length;
						break;
					}

				if (numberStrands == -1)
					break;   
				numberStrands += w.getStrands().length;
				if (numberStrands < minNumberStrands) {
					windowId = i;
					minNumberStrands = numberStrands;
					nextWindowId = j;
				}
			}
			if (windowId == -1)
				break;
			failedToMerge[windowId] = true;

			if (!isToMergeGivenParentSolution(windowId, nextWindowId) || !isToMergeGivenWindowSize( windowId, nextWindowId))
				continue;

			Util.pln(
				"Trying to merge window "
					+ windowId
					+ " ("
					+ windows[windowId].getStrands().length
					+ ") with "
					+ nextWindowId
					+ " ("
					+ windows[nextWindowId].getStrands().length
					+ ")");

			try {
				windows[nextWindowId].mergeWithPrevious(
					windows[windowId],
					computeMinMissingAround(windowId, nextWindowId),
					computeMinScoreAround(windowId, nextWindowId));
				windows[windowId] = null;
			} catch (OutOfMemoryError e) {
				Util.pln("Gave up out of memory");
				continue;
			} catch (MaxSizeExceeded e) {
				Util.pln("MaxSizeExceeded");
				continue;
			}
		}
		windows = (Window[]) Util.stripEmpties(windows);
		
		summary = null;
	}

    //--------produce a subsolution by fixing a spin system at a position--------------
    void fixUniqueAlternative(int spinSystemId, int pos) throws DeadEnd {

        for (int i = 0; i < windows.length; i++) {
            windows[i].fixUniqueAlternative(spinSystemId, pos);
        }

        fixedPos.set(pos);
        
        summary = null;
    }

	//-----------------------------------------------------------------------
	//------------------------------METHODS----------------------------------
	//-----------------------------------------------------------------------
	/**
	 * A summary gives, for each position in the protein, the ids of the spin
	 * systems occuring at that position.
	 * 
	 * This implementation tries to save work by finding a window in the parent
	 * that has the same starting position, length and size and then sharing the
	 * parent's summary.
	 * 
	 * It also only recomputes the summary if the Solution has changed since
	 * the summary was last computed.
	 */
	public void computeSummary() {
		if (summary == null) {
			summary = new BitSet[protein.getLength()];
			int currentWindowIndex = 0;
	
			for (int pos = 0; pos < protein.getLength(); pos++) {
				if (windows[currentWindowIndex].getPosRight() < pos) // JV: when do we need this ?
					currentWindowIndex++;
				summary[pos] = windows[currentWindowIndex].computeSummary(pos);
			}
		}
	}

	/**
	 * Find the position with the smallest cardinality.
	 * @return a valid position in the protein or -1 if all positions are fixed.
	 */
	public int findBestDefinedPosition() {

		int bestDefinedPos = -1;
		int minCardinality = Integer.MAX_VALUE;

		computeSummary();
		for (int pos = 0; pos < protein.getLength(); pos++) {

			if (protein.getAssignablePositions().test(pos) == false || fixedPos.test(pos) == true)
				continue;

			int cardinality = summary[pos].cardinality();
			if (cardinality < minCardinality) {
				minCardinality = cardinality;
				bestDefinedPos = pos;
			}
			if (cardinality == 1)
				return pos;
		}

		return bestDefinedPos;
	}

	public int[] findIdToFix(int pos) {

		// determine the ids
		int numberCandidates = summary[pos].cardinality();
		int[] candidateIds = new int[numberCandidates];

		for (int cursor = 0, i = 0; i < protein.getSpinSystems().length; i++) {
			if (summary[pos].test(i))
				candidateIds[cursor++] = protein.getSpinSystems()[i].getId();
			if (cursor == numberCandidates)
				break;
		}

		if (numberCandidates == 1)
			return candidateIds;

		// determine the order
		double metric[] = new double[numberCandidates];

		for (int i = 0; i < numberCandidates; i++) {
			int currentWindowIndex = 0;

			for (int testPos = 0; testPos < protein.getLength(); testPos++) {
				if (windows[currentWindowIndex].getPosRight() < testPos)
					currentWindowIndex++;

				if (windows[currentWindowIndex].getPosLeft() <= pos
					&& pos <= windows[currentWindowIndex].getPosRight()) {
					metric[i] = windows[currentWindowIndex].findMinPenalizedScore(candidateIds[i], testPos);
					break;
				}
			}
		}

		// sort the candidates
		Pair[] toSort = new Pair[numberCandidates];
		for (int i = 0; i < numberCandidates; i++)
			toSort[i] = new Pair(new Integer(candidateIds[i]), metric[i]);

		Arrays.sort(toSort);
		for (int i = 0; i < numberCandidates; i++)
			candidateIds[i] = ((Integer) toSort[i].getObject()).intValue();

		return candidateIds;
	}

	public void filter() throws DeadEnd {

		summary = null;

		// window - wise
		int minMissings = 0;
		for (int i = 0; i < windows.length; i++)
			minMissings += windows[i].getMinMissings();
		if (minMissings > Worker.maxMissings)
			throw new DeadEnd();

		double minScore = 0;
		for (int i = 0; i < windows.length; i++)
			minScore += windows[i].getMinScore();
		if (minScore > Worker.mappingsFound.getMaxPenalizedScore() + Worker.deltaScore)
			throw new DeadEnd();

		// strand - wise
		for (int i = 0; i < windows.length; i++)
			windows[i].filter(minMissings - windows[i].getMinMissings(), minScore - windows[i].getMinScore());
	}

	private boolean isToMergeGivenParentSolution(int i, int j) {

		if (parentWindowInfo == null)
			return true;

		int id = findWindowIdInParentSolution(i);
		int idNext = findWindowIdInParentSolution(j);
		if (id != -1
			&& idNext != -1
			&& ((parentWindowInfo[id].getNumStrands() * 0.9) - windows[i].getStrands().length
				> 100)
			&& ((parentWindowInfo[idNext].getNumStrands() * 0.9) - windows[j].getStrands().length
				> 100)
			&& windows[i].getStrands().length > 500
			&& windows[j].getStrands().length > 500) {
			return false;
		}
		return true;
	}

	private boolean isToMergeGivenWindowSize( int i, int j) {
		if (windows[i].getStrands().length > 100
			&& windows[j].getStrands().length > 100
			&& windows[i].getStrands().length + windows[j].getStrands().length > Worker.maxWindowSize / 10) {
			return false;
		}
		return true;
	}

	/**
	 * Given an index in the window array
	 * @param ii
	 * @return a valid position or -1
	 */
	private int findWindowIdInParentSolution(int currentWindowIndex) {
		for (int i = 0; i < parentWindowInfo.length; i++)
			if (parentWindowInfo[i].getPosLeft() == windows[currentWindowIndex].getPosLeft()
				&& parentWindowInfo[i].getPosRight() == windows[currentWindowIndex].getPosRight())
				return i;
		return -1;
	}

	private double computeMinScoreAround(int i, int next) {
		double minScoreAround = 0;
		for (int j = 0; j < windows.length; j++)
			if (j != i && j != next && windows[j] != null)
				minScoreAround += windows[j].getMinScore();
		return minScoreAround;
	}
	private int computeMinMissingAround(int i, int next) {
		int minMissingsAround = 0;
		for (int j = 0; j < windows.length; j++)
			if (j != i && j != next && windows[j] != null)
				minMissingsAround += windows[j].getMinMissings();
		return minMissingsAround;
	}

	public String toStringByWindows() {
		String s = "";
		for (int i = 0; i < windows.length; i++) {
			s += "["
				+ Util.pad(windows[i].getPosLeft() + "", 3)
				+ ","
				+ Util.pad(windows[i].getPosRight() + "", 3)
				+ "] * "
				+ Util.pad("" + windows[i].getStrands().length, 5)
				+ "   Min missings "
				+ windows[i].getMinMissings();

			if (!windows[i].containsReference())
				s += "   Reference strand excluded";
			s += "\n";
		}
		return s;
	}

	public String toStringByPositions() {

		computeSummary();

		String s = "";
		for (int pos = 0; pos < protein.getLength(); pos++) {
			s += protein.getResidueAt(pos).getShortName()
				+ "  ["
				+ Util.pad(pos + "", 3)
				+ "] * "
				+ Util.pad("" + summary[pos].cardinality(), 6);
			s += "\n";
		}
		return s;
	}

	//-----------------------------------------------------------------------
	//------------------------------ACCESSORS--------------------------------
	//-----------------------------------------------------------------------
	public Window[] getWindows() {
		return windows;
	}

	public BitSet getFixedPos() {
		return fixedPos;
	}
	
	public static Solution getRootSolution() {
		return root;
	}

	//-----------------------------------------------------------------------
	//----------------------------EXCEPTIONS---------------------------------
	//-----------------------------------------------------------------------
	public static class DeadEnd extends Exception {
	}

	public static class MaxSizeExceeded extends Exception {
	}
	
	//-----------------------------------------------------------------------
	//------------------------------DEBUG------------------------------------
	//-----------------------------------------------------------------------
	protected void finalize() {
		instanceCount.decrement();	
	}

	public static int getInstanceCount() {
		return instanceCount.get();
	}
	
	//-----------------------------------------------------------------------
	//-----------------------------CLASSES-----------------------------------
	//-----------------------------------------------------------------------
	
	// The only information about the parent solution that is needed is the
	// edges of its windows and how many strands each window had.  Storing
	// only this information about the parent eliminates the need to keep old
	// solutions around just because their subsolutions are still live.
	private static class WindowInfo implements Serializable {
		final private int posLeft;
		final private int posRight;
		final private int numStrands;
		
		public WindowInfo(Window window) {
			posLeft = window.getPosLeft();
			posRight = window.getPosRight();
			numStrands = window.getStrands().length;
		}
		
		public int getPosLeft() {
			return posLeft;
		}
		
		public int getPosRight() {
			return posRight;
		}
		
		public int getNumStrands() {
			return numStrands;
		}
	}

	//-----------------------------------------------------------------------
	//----------------------------CLASSES------------------------------------
	//-----------------------------------------------------------------------

	// Reconstruct a Solution given the fixed spin systems and
	// information about the parent solution's windows.  Reconstructing
	// solutions eliminates the need to pass them directly to remote systems,
	// dramatically reducing network usage.
	public static class Reconstructor implements Serializable {

		final private int[] spinSystemIds;
		final private int[] positions;
		final private WindowInfo[] parentWindowInfo;

		public Reconstructor(Solution solution) {
			if (solution == null) {
				spinSystemIds = null;
				positions = null;
				parentWindowInfo = null;
				return;
			}

			parentWindowInfo = solution.parentWindowInfo;
			spinSystemIds = new int[solution.fixedPos.cardinality()];
			positions = new int[spinSystemIds.length];

			solution.computeSummary();
			int i = 0;
			for (int j = 0; j < protein.getLength(); j++) {
				if (solution.fixedPos.test(j)) {
					spinSystemIds[i] = solution.summary[j].firstSet();
					positions[i] = j;
					i++;
				}
			}
		}

		public Solution get() throws DeadEnd {
			if (spinSystemIds == null)
				return root;
			Util.pln("Reconstructing a solution with " + positions.length + " fixed positions.");
			Solution result = new Solution(spinSystemIds, positions, parentWindowInfo);
			result.filter();
			result.checkAndMergeWindows();
			return result;
		}
	}
	
	// Encapsulates a soft reference to a Solution.  This allows the garbage collector
	// to collect the Solution if memory is running low; if the Solution is needed after
	// being collected, it will be reconstructed on demand.
	public static class Reference {
		private SoftReference reference;
		private Reconstructor reconstructor;
		
		public Reference(Solution solution) {
			this.reconstructor = new Reconstructor(solution);
			this.reference = new SoftReference(solution);
		}
		
		public Reference(Reconstructor reconstructor) throws DeadEnd {
			this.reconstructor = reconstructor;
			this.reference = new SoftReference(null);
		}
		
		public Solution get() throws DeadEnd {
			Solution result = (Solution) reference.get();
			if (result == null) {
				result = reconstructor.get();
				reference = new SoftReference(result);
			}
			return result;
		}
		
		public Reconstructor reconstructor() {
			return reconstructor;
		}
		
		// Call this after modifying the Solution to update the reconstructor, while
		// your hard reference to the Solution still exists.
		public void solutionChanged() {
			Solution solution = (Solution) reference.get();
			if (solution != null) {
				reconstructor = new Reconstructor(solution);
			}
		}
	}
}
