package assign;

import input.*;
import input.Protein;
import input.SpinSystem;
import util.BitSet;
import util.ChiSquareDistribution;
import util.Util;

import java.io.Serializable;
import java.util.Arrays;

public final class Strand implements Serializable {

	private double scoreMatch, scoreAlign, scoreJoint;
	private int dfMatch, dfAlign, dfJoint;

	private SpinSystem left, right;
	private int posLeft, posRight;
	private BitSet spinSystemCoverage;
	private int numberMissings;
	private SpinSystem[] spinSystems;
	private boolean isReference;

	private static Protein protein = Worker.protein;
	private static ScoreHandler scoreHandler = Worker.scoreHandler;

	//-----------------------------------------------------------------------------------
	//-----------------------------CONSTRUCTORS------------------------------------------
	//-----------------------------------------------------------------------------------

	//----------------from a spin system at a position (assume valid)-------------------- 
	public Strand(SpinSystem spinSystem, int pos) {

		scoreMatch = 0;
		dfMatch = 0;
		scoreAlign = 0;
		dfAlign = 0;
		// last position - complete the score
		if (pos == protein.getLastPos()) {
			scoreAlign += scoreHandler.getScoreAlign(pos, spinSystem, protein.getEmptySpinSystem());
			dfAlign += scoreHandler.getDfAlign(spinSystem, protein.getEmptySpinSystem());
		}
		scoreJoint = scoreAlign;
		dfJoint = dfAlign;

		right = left = spinSystem;
		posLeft = posRight = pos;

		spinSystemCoverage = new BitSet(0, protein.getSpinSystems().length);
		if (!spinSystem.isEmpty())
			spinSystemCoverage.set(spinSystem.getId());

		numberMissings = 0;
		if (spinSystem.isEmpty() && protein.getAssignablePositions().test(pos) == true)
			numberMissings = 1;

		spinSystems = new SpinSystem[] { spinSystem };

		isReference = true;
		if (protein.getAssignedInReference().test(pos) == true && spinSystem.getReferenceId() != pos)
			isReference = false;
		if (protein.getAssignedInReference().test(pos) == false && !spinSystem.isEmpty())
			isReference = false;


	}

	//------------------------from two strands (assume valid)------------------------ 
	public Strand(Strand strandA, Strand strandB) {

		// score match
		scoreMatch = strandA.getScoreMatch() + strandB.getScoreMatch();
		scoreMatch += scoreHandler.getScoreMatch(strandA.getRight(), strandB.getLeft());

		dfMatch = strandA.getDfMatch() + strandB.getDfMatch();
		dfMatch += scoreHandler.getDfMatch(strandA.getRight(), strandB.getLeft());

		// score align
		scoreAlign = strandA.getScoreAlign() + strandB.getScoreAlign();
		scoreAlign
			+= scoreHandler.getScoreAlign(strandA.getPosRight(), strandA.getRight(), strandB.getLeft());
		dfAlign = strandA.getDfAlign() + strandB.getDfAlign();
		dfAlign += scoreHandler.getDfAlign(strandA.getRight(), strandB.getLeft());

		// score joint
		scoreJoint = strandA.getScoreJoint() + strandB.getScoreJoint();
		scoreJoint
			+= scoreHandler.getScoreJoint(strandA.getPosRight(), strandA.getRight(), strandB.getLeft());
		dfJoint = strandA.getDfJoint() + strandB.getDfJoint();
		dfJoint += scoreHandler.getDfJoint(strandA.getRight(), strandB.getLeft());

		// other fields
		left = strandA.getLeft();
		right = strandB.getRight();

		posLeft = strandA.getPosLeft();
		posRight = strandB.getPosRight();

		spinSystemCoverage = strandA.getSpinSystemCoverage().copy();
		spinSystemCoverage.union(strandB.getSpinSystemCoverage());

		numberMissings = strandA.getNumberMissingSpinSystems() + strandB.getNumberMissingSpinSystems();
		spinSystems = (SpinSystem[]) Util.concat(strandA.spinSystems, strandB.spinSystems);

		isReference = strandA.isReference() && strandB.isReference();

	}

	//------------------------from two strands with checks--------------------------- 
	public static Strand CheckAndMake(
		Strand strandA,
		Strand strandB,
		int minMissingsAround,
		double minScoreJointAround) {

		// score match
		double outScoreMatch = strandA.getScoreMatch() + strandB.getScoreMatch();
		outScoreMatch += scoreHandler.getScoreMatch(strandA.getRight(), strandB.getLeft());

		int outDfMatch = strandA.getDfMatch() + strandB.getDfMatch();
		outDfMatch += scoreHandler.getDfMatch(strandA.getRight(), strandB.getLeft());

		// score align
		double outScoreAlign = strandA.getScoreAlign() + strandB.getScoreAlign();
		outScoreAlign
			+= scoreHandler.getScoreAlign(strandA.getPosRight(), strandA.getRight(), strandB.getLeft());

		int outDfAlign = strandA.getDfAlign() + strandB.getDfAlign();
		outDfAlign += scoreHandler.getDfAlign(strandA.getRight(), strandB.getLeft());

		// score joint
		double outScoreJoint = strandA.getScoreJoint() + strandB.getScoreJoint();
		outScoreJoint
			+= scoreHandler.getScoreJoint(strandA.getPosRight(), strandA.getRight(), strandB.getLeft());

		int outDfJoint = strandA.getDfJoint() + strandB.getDfJoint();
		outDfJoint += scoreHandler.getDfJoint(strandA.getRight(), strandB.getLeft());

		// number of missings
		int outNumberMissings = strandA.getNumberMissingSpinSystems() + strandB.getNumberMissingSpinSystems();

		// ------------------------is to make?---------------------------------------
		if (strandB.getPosRight() - strandA.getPosLeft() > 1) {

			if (outScoreMatch > 0
				&& outScoreMatch > ChiSquareDistribution.quantile(1 - Worker.beta, outDfMatch)) {
				if (strandA.isReference() && strandB.isReference())
					Util.pln(
						"Reject reference strand ["
							+ strandA.getPosLeft()
							+ ","
							+ strandB.getPosRight()
							+ "]. Score match " + outScoreMatch + ", df " + outDfMatch);
				return null;
			}

			if (outScoreAlign > 0
				&& outScoreAlign > ChiSquareDistribution.quantile(1 - Worker.beta, outDfAlign)) {
				if (strandA.isReference() && strandB.isReference())
					Util.pln(
						"Reject reference strand ["
							+ strandA.getPosLeft()
							+ ","
							+ strandB.getPosRight()
							+ "]. Score align " + outScoreAlign + ", df " + outDfAlign);
				return null;
			}

			if (outScoreJoint > 0
				&& outScoreJoint > ChiSquareDistribution.quantile(1 - Worker.beta, outDfJoint)) {
				if (strandA.isReference() && strandB.isReference())
					Util.pln(
						"Reject reference strand ["
							+ strandA.getPosLeft()
							+ ","
							+ strandB.getPosRight()
							+ "]. Score joint " + outScoreJoint + ", df " + outDfJoint);
				return null;
			}

			if (!isConsistentWithOtherWindows(outScoreJoint,
				outNumberMissings,
				minMissingsAround,
				minScoreJointAround)) {
				if (strandA.isReference() && strandB.isReference())
					Util.pln(
						"Reject reference strand ["
							+ strandA.getPosLeft()
							+ ","
							+ strandB.getPosRight()
							+ "] - inconsistent with other windows");
				return null;
			}
		}

		return new Strand(
			strandA,
			strandB,
			outScoreMatch,
			outDfMatch,
			outScoreAlign,
			outDfAlign,
			outScoreJoint,
			outDfJoint,
			outNumberMissings);
	}

	//-----------------------------pro forma constructor--------------------------- 
	private Strand(
		Strand strandA,
		Strand strandB,
		double scoreMatch,
		int dfMatch,
		double scoreAlign,
		int dfAlign,
		double scoreJoint,
		int dfJoint,
		int numberMissings) {

		// scores and dfs
		this.scoreMatch = scoreMatch;
		this.dfMatch = dfMatch;
		this.scoreAlign = scoreAlign;
		this.dfAlign = dfAlign;
		this.scoreJoint = scoreJoint;
		this.dfJoint = dfJoint;
		this.numberMissings = numberMissings;

		// other fields
		left = strandA.getLeft();
		right = strandB.getRight();

		posLeft = strandA.getPosLeft();
		posRight = strandB.getPosRight();

		spinSystemCoverage = strandA.getSpinSystemCoverage().copy();
		spinSystemCoverage.union(strandB.getSpinSystemCoverage());
		spinSystems = (SpinSystem[]) Util.concat(strandA.spinSystems, strandB.spinSystems);

		isReference = strandA.isReference() && strandB.isReference();

	}

	//-----------------------------------------------------------------------------------
	//----------------------------CONSISTENCY CHECKS-------------------------------------
	//-----------------------------------------------------------------------------------

	//--------------------strand of length 1 (consistency)--------------------- 
	public static boolean isConsistent(SpinSystem spinSystem, int pos) {

		//invalid spin system at the position
		if (protein.getAssignablePositions().test(pos) == false && !spinSystem.isEmpty())
			return false;
		if (protein.getResidueAt(pos).isGlycine() && !spinSystem.canBeGly())
			return false;
		if (!protein.getResidueAt(pos).isGlycine() && !spinSystem.canBeNonGly())
			return false;

		//invalid spin system following a position
		if (pos > 0) {
			if (protein.getResidueAt(pos - 1).isGlycine() && !spinSystem.canFollowGly())
				return false;
			if (!protein.getResidueAt(pos - 1).isGlycine() && !spinSystem.canFollowNonGly())
				return false;
			//after an unassignable position - check the score
			if (protein.getAssignablePositions().test(pos - 1) == false)
				if (!scoreHandler.isValidAlign(pos - 1, protein.getEmptySpinSystem(), spinSystem))
					return false;
		}

		//last position or before an unassignable position - check the score
		if (pos == protein.getLastPos() || protein.getAssignablePositions().test(pos + 1) == false)
			if (!scoreHandler.isValidAlign(pos, spinSystem, protein.getEmptySpinSystem()))
				return false;

		return true;
	}

	//--------------------strand of length > 1 (assume valid position)----------------- 
	public static boolean isConsistent(Strand strandA, Strand strandB) {

		// same spin system in two strands
		if (!strandA.spinSystemCoverage.disjoint(strandB.spinSystemCoverage))
			return false;

		// invalid match score
		if (!scoreHandler.isValidMatch(strandA.getRight(), strandB.getLeft()))
			return false;

		// invalid align score
		if (!scoreHandler.isValidAlign(strandA.getPosRight(), strandA.getRight(), strandB.getLeft()))
			return false;

		// invalid joint score
		if (!scoreHandler.isValidJoint(strandA.getPosRight(), strandA.getRight(), strandB.getLeft()))
			return false;

		return true;
	}

	//--------------------------solution-wise consistency--------------------------
	public static boolean isConsistentWithOtherWindows(
		double scoreJoint,
		int numberMissings,
		int minMissingsAround,
		double minScoreJointAround) {

		if (numberMissings + minMissingsAround > Worker.maxMissings)
			return false;
		if (scoreJoint + minScoreJointAround > Worker.mappingsFound.getMaxPenalizedScore())
			return false;

		return true;
	}

	//-----------------------------------------------------------------------------------
	//-------------------------------------METHODS---------------------------------------
	//-----------------------------------------------------------------------------------
	public String toString() {

		String s = "";
		for (int i = 0; i < spinSystems.length - 1; i++)
			s += spinSystems[i].toString() + ", ";
		s += spinSystems[spinSystems.length - 1].toString() + "";

		return s;
	}
	
	//-----------------------------------------------------------------------------------
	//---------------------------------OBJECT IDENTITY-----------------------------------
	//-----------------------------------------------------------------------------------
	
	// We only worry about the structure of the strand, not its scoring information or
	// computed data.
	
	public boolean equals(Object o) {
		Strand s = (Strand) o;
		if (posLeft != s.posLeft)
			return false;
		if (posRight != s.posRight)
			return false;
		if (numberMissings != s.numberMissings)
			return false;
		return Arrays.equals(spinSystems, s.spinSystems);
	}
	
	public int hashCode() {
		int result = 0;
		for (int i = 0; i < spinSystems.length; i++)
			result ^= spinSystems[i].hashCode();
		return result ^ posLeft ^ posRight ^ numberMissings;
	}
	
	//-----------------------------------------------------------------------------------
	//------------------------------------ACCESSORS--------------------------------------
	//-----------------------------------------------------------------------------------
	public double getScoreAlign() {
		return scoreAlign;
	}

	public int getDfAlign() {
		return dfAlign;
	}

	public double getScoreMatch() {
		return scoreMatch;
	}

	public int getDfMatch() {
		return dfMatch;
	}

	public double getScoreJoint() {
		return scoreJoint;
	}

	public int getDfJoint() {
		return dfJoint;
	}

	public SpinSystem getLeft() {
		return left;
	}

	public SpinSystem getRight() {
		return right;
	}

	public int getPosRight() {
		return posRight;
	}

	public int getPosLeft() {
		return posLeft;
	}

	public BitSet getSpinSystemCoverage() {
		return spinSystemCoverage;
	}

	public int getNumberMissingSpinSystems() {
		return numberMissings;
	}

	public SpinSystem[] getSpinSystems() {
		return spinSystems;
	}

	public int getSpinSystemIdAt(int pos) {
		return spinSystems[pos - posLeft].getId();
	}

	public boolean isReference() {
		return isReference;
	}

}
