/*
 * Created on Dec 26, 2003
 *
 */
package assign;

import util.BitSet;
import util.ChiSquareDistribution;
import util.Util;

public class Window {

	private Strand[] strands;
	private int posLeft, posRight;

	private int minMissings;
	private double minScore;
	private boolean containsReference;

	//-----------------------------------------------------------------------------------
	//-----------------------------CONSTRUCTORS------------------------------------------
	//-----------------------------------------------------------------------------------
	//---------------------from spin systems at a position-------------------------------
	public Window(int pos) {

		strands = new Strand[Worker.protein.getSpinSystems().length];
		posLeft = pos;
		posRight = pos;
		containsReference = false;

		int cursor = 0;
		for (int i = 0; i < Worker.protein.getSpinSystems().length; i++) {

			if (Strand.isConsistent(Worker.protein.getSpinSystems()[i], pos)) {
				strands[cursor] = new Strand(Worker.protein.getSpinSystems()[i], pos);

				if (strands[cursor].isReference())
					containsReference = true;
				cursor++;
			}
		}

		if (cursor == 0)
			Util.error("Error initializing a window at position " + pos + ": no candidate mappings.");
		strands = (Strand[]) Util.truncate(strands, cursor);
		minMissings = computeMinMissings();
		minScore = computeMinScore();
	}

	//--------------from a window by fixing a spin system at a position--------------
	public Window(Window window, int spinSystemId, int pos) throws Solution.DeadEnd {
		this(window, new int[] { spinSystemId }, new int[] { pos });
	}
	
	//---------from a window by fixing multiple spin systems at various positions----
	public Window(Window window, int[] spinSystemIds, int[] positions) throws Solution.DeadEnd {
		strands = (Strand[]) Util.copy(window.getStrands());
		posLeft = window.getPosLeft();
		posRight = window.getPosRight();
		containsReference = window.containsReference();
		
		for (int i = 0; i < positions.length; i++) {

			// position belongs to this window
			if (posLeft <= positions[i] && positions[i] <= posRight) {
				for (int j = 0; j < strands.length; j++) {
	
					if (strands[j] != null && strands[j].getSpinSystemIdAt(positions[i]) != spinSystemIds[i]) {
	
						if (strands[j].isReference())
							containsReference = false;
						strands[j] = null;
					}
				}
			}
	
			// position does not belong to this window
			if (positions[i] < posLeft || posRight < positions[i]) {
				for (int j = 0; j < strands.length; j++) {
					if (strands[j] != null && strands[j].getSpinSystemCoverage().test(spinSystemIds[i])) {
						if (strands[j].isReference())
							containsReference = false;
						strands[j] = null;
					}
				}
			}
		}

		if (strands.length == 0)
			throw new Solution.DeadEnd();

		strands = (Strand[]) Util.stripEmpties(strands);
		minMissings = computeMinMissings();
		minScore = computeMinScore();
	}

	//---------------------from a window by fixing a strand---------------------
	public Window(Window window, Strand strandToFix) throws Solution.DeadEnd {

		posLeft = window.getPosLeft();
		posRight = window.getPosRight();
		containsReference = window.containsReference();

		// strand belongs to this window
		if (posLeft == strandToFix.getPosLeft() && posRight == strandToFix.getPosRight()) {

			strands = new Strand[] { strandToFix };
			if (!strandToFix.isReference())
				containsReference = false;
		}

		// position does not belong to this window
		if (posLeft != strandToFix.getPosLeft() || posRight != strandToFix.getPosRight()) {

			strands = (Strand[]) Util.copy(window.getStrands());
			for (int i = 0; i < strands.length; i++) {
				if (!strands[i].getSpinSystemCoverage().disjoint(strandToFix.getSpinSystemCoverage())) {
					if (strands[i].isReference())
						containsReference = false;
					strands[i] = null;
					break;
				}
			}
		}

		if (strands.length == 0)
			throw new Solution.DeadEnd();

		strands = (Strand[]) Util.stripEmpties(strands);
		minMissings = computeMinMissings();
		minScore = computeMinScore();

	}

	//-----------------------------------------------------------------------
	//------------------------------MUTATORS----------------------------------
	//-----------------------------------------------------------------------
	//-----------------------------filtering a window--------------------------------
	public void filter(int minMissingsAround, double minScoreAround) throws Solution.DeadEnd {

		for (int i = 0; i < strands.length; i++) {

			if (!Strand
				.isConsistentWithOtherWindows(
					strands[i].getScoreJoint(),
					strands[i].getNumberMissingSpinSystems(),
					minMissingsAround,
					minScoreAround)) {
				if (strands[i].isReference())
					containsReference = false;
				strands[i] = null;
			}
		}
		strands = (Strand[]) Util.stripEmpties(strands);
		if (strands.length == 0)
			throw new Solution.DeadEnd();

	}

	//--------------------merge previous (left) window with this (right) window-------------------
	public void mergeWithPrevious(Window window, int minMissingsAround, double minScoreAround)
		throws Solution.DeadEnd, Solution.MaxSizeExceeded {

		Strand[] strandsA = window.getStrands();
		Strand[] strandsB = strands;

		int outStrandsLength = Math.min(strandsA.length * strandsB.length, Worker.maxWindowSize);
		if (outStrandsLength < 0)
			outStrandsLength = Worker.maxWindowSize;
		Strand[] outStrands = new Strand[outStrandsLength];

		boolean outContainsReference = false;

		int cursor = 0;
		for (int cursorB = 0; cursorB < strandsB.length; cursorB++) {

			Strand strandB = strandsB[cursorB];
			int countMatchB = 0;

			for (int cursorA = 0; cursorA < strandsA.length; cursorA++) {
				if (cursor >= outStrandsLength) {
					strands = (Strand[]) Util.stripEmpties(strands);
					throw new Solution.MaxSizeExceeded();
				}

				Strand strandA = strandsA[cursorA];
				if (Strand.isConsistent(strandA, strandB)) {
					outStrands[cursor] =
						Strand.CheckAndMake(strandA, strandB, minMissingsAround, minScoreAround);

					if (outStrands[cursor] != null) {
						if (outStrands[cursor].isReference())
							outContainsReference = true;
						cursor++;
						countMatchB++;
					}
				}
			}

			// remove the strand from the current window for which no matches has been found
			if (countMatchB == 0)
				strandB = null;
		}

		if (cursor == 0)
			throw new Solution.DeadEnd();
		Util.p(toString(window, cursor));

		strands = (Strand[]) Util.truncate(outStrands, cursor);
		containsReference = outContainsReference;
		posLeft = window.getPosLeft();
		minMissings = computeMinMissings();
		minScore = computeMinScore();
		Util.pln(minMissings);
	}

    //--------------fixing a single aternative at a position--------------
    void fixUniqueAlternative(int spinSystemId, int pos) throws Solution.DeadEnd {

        // position belongs to this window
        if (posLeft <= pos && pos <= posRight) {
            for (int i = 0; i < strands.length; i++) {

                if (strands[i].getSpinSystemIdAt(pos) != spinSystemId) {

                    if (strands[i].isReference())
                        containsReference = false;
                    strands[i] = null;
                }
            }
        }

        // position does not belong to this window
        if (pos < posLeft || posRight < pos) {
            for (int i = 0; i < strands.length; i++) {
                if (strands[i].getSpinSystemCoverage().test(spinSystemId)) {
                    if (strands[i].isReference())
                        containsReference = false;
                    strands[i] = null;
                }
            }
        }

        if (strands.length == 0)
            throw new Solution.DeadEnd();

        strands = (Strand[]) Util.stripEmpties(strands);
        minMissings = computeMinMissings();
        minScore = computeMinScore();

    }
	//-----------------------------------------------------------------------
	//------------------------------METHODS----------------------------------
	//-----------------------------------------------------------------------
	public BitSet computeSummary(int pos) {
		BitSet summary = new BitSet(0, Worker.protein.getSpinSystems().length);
		for (int j = 0; j < strands.length; j++) {
			summary.set(strands[j].getSpinSystemIdAt(pos));
		}
		return summary;
	}

	public double findMinPenalizedScore(int id, int pos) {

		if (id == Worker.protein.getEmptySpinSystem().getId())
			return Double.MAX_VALUE;

		int maxDfMatch = computeMaxDfMatch();
		double result = Double.MAX_VALUE;
		for (int i = 0; i < strands.length; i++) {
			if (strands[i].getSpinSystemIdAt(pos) == id) {

				double penalizedScore = strands[i].getScoreMatch();
				int penaltyDf = maxDfMatch - strands[i].getDfMatch();

				if (penaltyDf > 0)
					penalizedScore += ChiSquareDistribution.quantile(1 - Worker.alpha, penaltyDf);
				result = Math.min(penalizedScore, result);

			}
		}
		return result;
	}

	private int computeMaxDfMatch() {

		int maxDfMatch = Integer.MIN_VALUE;
		for (int i = 0; i < strands.length; i++)
			maxDfMatch = Math.max(strands[i].getDfMatch(), maxDfMatch);
		return maxDfMatch;
	}

	private int computeMinMissings() {

		int count = Integer.MAX_VALUE;
		for (int i = 0; i < strands.length; i++)
			count = Math.min(strands[i].getNumberMissingSpinSystems(), count);
		return count;
	}

	private double computeMinScore() {

		double score = Integer.MAX_VALUE;
		for (int i = 0; i < strands.length; i++)
			score = Math.min(strands[i].getScoreJoint(), score);
		return score;
	}

	private String toString(Window w, int outLength) {
		return "Merged"
			+ "["
			+ Util.pad(w.getPosLeft() + "", 3)
			+ ","
			+ Util.pad(w.getPosRight() + "", 3)
			+ "] * "
			+ Util.pad("" + w.getStrands().length, 6)
			+ "  with  ["
			+ Util.pad(posLeft + "", 3)
			+ ","
			+ Util.pad(posRight + "", 3)
			+ "] * "
			+ Util.pad("" + strands.length, 6)
			+ "  -->  "
			+ Util.pad("" + outLength, 6)
			+ "   Min missings ";
	}

	//-----------------------------------------------------------------------------------
	//------------------------------------ACCESSORS--------------------------------------
	//-----------------------------------------------------------------------------------
	public Strand[] getStrands() {
		return strands;
	}

	public int getPosRight() {
		return posRight;
	}

	public int getPosLeft() {
		return posLeft;
	}

	public int getMinMissings() {
		return minMissings;
	}

	public double getMinScore() {
		return minScore;
	}

	public boolean containsReference() {
		return containsReference;
	}
}
