package input;

import java.io.BufferedReader;
import java.io.FileNotFoundException;
import java.io.FileOutputStream;
import java.io.FileReader;
import java.io.IOException;
import java.io.PrintWriter;
import java.util.Random;
import java.util.StringTokenizer;

import util.Util;
import assign.Worker;

public class DataGenerator {

    // original data
    String comment = "";
    String sequence = "";
    double[][] trueChemicalShifts;

    // simulated data
    Random random = new Random();

    //-----------------------------------------------------------------------------------------
    //------------------------------------CONSTRUCTORS-----------------------------------------
    //-----------------------------------------------------------------------------------------

    //---------- takes data in the bmrb format and writes into the mba and txt formats---------
    public DataGenerator(String inFileREFDB, String outFileTxt, String outFileMBA_full, String outFileMBA, double hnv, double nv, double cov, double cav, double cbv, double hav) {

        // read the BMRB file and memorise
        readTrueDataREFDBformat(inFileREFDB);
        Util.pln("The original data has " + sequence.length() + " positions.");

        // generate simulated chemical shifts
        if (!outFileMBA.equals(""))
            simulateMissingChemicalShifts();

        // write the simulated data in the readable format
        writeTrueChemicalShiftsTxtFormat(outFileTxt);

        //write the simulated spin systems into the outFile in the MBA format
        if (outFileMBA.equals(""))
            return;
        writeSimSpinSystemsMBAformat(outFileMBA_full, outFileMBA, hnv, nv, cov, cav, cbv, hav);

    }

    //----------------------- takes data in the txt format and memorizes-------------------------------------
    public DataGenerator(String inFileTXT) {

        trueChemicalShifts = new double[Worker.protein.getLength()][8];
        try {
            BufferedReader inputStream = new BufferedReader(new FileReader(inFileTXT));
            String line = inputStream.readLine();
            int pos = 0;
            while (line != null) {

                StringTokenizer tokenizer = new StringTokenizer(line, ",");
                tokenizer.nextToken();
                for (int j = 0; j < 8; j++)
                    trueChemicalShifts[pos][j] = Double.parseDouble(tokenizer.nextToken().trim());
                pos++;
                line = inputStream.readLine();
            }
            inputStream.close();

        } catch (FileNotFoundException e) {
            Util.pln("File " + inFileTXT + " not found. The program will now stop");
            System.exit(0);
        } catch (IOException e) {
            Util.pln("Error reading from file " + inFileTXT + ". The program will now stop");
            System.exit(0);
        }

    }

    //-----------------------------------------------------------------------------------------
    //----------------------------------------METHODS------------------------------------------
    //-----------------------------------------------------------------------------------------

    //-----------------read the original data in REFDB format----------------------------------
    private void readTrueDataREFDBformat(String inFileBMRB) {
        try {
            BufferedReader inputStream = new BufferedReader(new FileReader(inFileBMRB));

            // comment and sequence
            comment = "Protein " + inputStream.readLine();
            sequence = inputStream.readLine();
            comment += ";   " + sequence.length() + " residues.";

            // original spin systems: array with columns id, HN, N15, CO, CA, CB, HA, HA2   
            trueChemicalShifts = new double[sequence.length()][8];
            for (int i = 0; i < sequence.length(); i++) {
                for (int j = 0; j < 8; j++)
                    trueChemicalShifts[i][j] = Double.NaN;
            }

            String line = inputStream.readLine();
            while (line != null) {

                StringTokenizer tokenizer = new StringTokenizer(line, " ");
                int pos = -1;
                int dim = -1;
                double chemicalShift = -1;
                for (int i = 0; i < 6; i++) {

                    if (i == 1)
                        pos = (int) Integer.parseInt(tokenizer.nextToken()) - 1;

                    if (i == 3) {
                        String dimName = tokenizer.nextToken();
                        if (dimName.equals("H"))
                            dim = 1;
                        if (dimName.equals("N"))
                            dim = 2;
                        if (dimName.equals("C"))
                            dim = 3;
                        if (dimName.equals("CA"))
                            dim = 4;
                        if (dimName.equals("CB"))
                            dim = 5;
                        if (dimName.equals("HA") || dimName.equals("HA2"))
                            dim = 6;
                        if (dimName.equals("HA3"))
                            dim = 7;
                    }

                    if (i == 5)
                        chemicalShift = (double) Double.parseDouble(tokenizer.nextToken());

                    if (i == 0 || i == 2 || i == 4)
                        tokenizer.nextToken();
                }

                if (pos != -1 && dim != -1) {
                    trueChemicalShifts[pos][0] = pos;
                    trueChemicalShifts[pos][dim] = chemicalShift;
                }

                line = inputStream.readLine();
            }
            inputStream.close();

        } catch (FileNotFoundException e) {
            Util.pln("File " + inFileBMRB + " not found. The program will now stop");
            System.exit(0);
        } catch (IOException e) {
            Util.pln("Error reading from file " + inFileBMRB + ". The program will now stop");
            System.exit(0);
        }

        // generate missing position ids
        for (int pos = 0; pos < sequence.length(); pos++)
            if (Double.isNaN(trueChemicalShifts[pos][0]))
                trueChemicalShifts[pos][0] = pos;

    }

    //---------------------------------generate simulated chemical shifts -------------------------------------------------------
    private void simulateMissingChemicalShifts() {

        String[] resTypes = { "H", "N", "C", "CA", "CB", "HA", "HA2" };
        for (int pos = 0; pos < sequence.length(); pos++) {

            // simulated residue
            Residue r = new Residue(sequence.substring(pos, pos + 1));
            r.simulate();
            for (int i = 1; i < 8; i++)
                if (Double.isNaN(trueChemicalShifts[pos][i]))
                    trueChemicalShifts[pos][i] = r.getSimChemicalShift(resTypes[i - 1]);
        }
    }

    //-------------------write the simulated chemical shifts in the readable format----------------------------------
    private void writeTrueChemicalShiftsTxtFormat(String outFileTxt) {
        PrintWriter outputStream = null;
        try {
            outputStream = new PrintWriter(new FileOutputStream(outFileTxt));
        } catch (FileNotFoundException e) {
            Util.pln("The output file for TXT format could not be created. The program will now stop");
            System.exit(0);
        }

        String s = "";
        for (int pos = 0; pos < sequence.length(); pos++) {
            s += sequence.substring(pos, pos + 1) + "," + Util.pad("" + ((int) trueChemicalShifts[pos][0]), 8);
            for (int j = 1; j < 8; j++)
                s += "," + Util.pad("" + Util.round(trueChemicalShifts[pos][j], 2), 8);
            s += "\n";
        }
        outputStream.print(s);
        outputStream.close();
    }

    //-------------------write the simulated spin systems into the outFile in the MBA format----------------------------------
    private void writeSimSpinSystemsMBAformat(String outFileMBA_full, String outFileMBA, double hv, double nv, double cov, double cav, double cbv, double hav) {

        String s_full = "";
        String s = "";

        int countMissingSS = 0;
        String simMissingSSat = "";

        int countExtraSS = 0;

        for (int pos = 1; pos < sequence.length(); pos++) {

            if (sequence.substring(pos, pos + 1).equals("P"))
                continue;

        // correct spin system
            String t =
                pos
                    + ","
                    + Util.round(trueChemicalShifts[pos][1] + random.nextGaussian() * Math.sqrt(hv), 2)
                    + ","
                    + Util.round(trueChemicalShifts[pos][2] + random.nextGaussian() * Math.sqrt(nv), 2)
                    + ",";
            // C-1                   
            if (random.nextDouble() > 0.006225374)
                t += Util.round(trueChemicalShifts[pos - 1][3] + random.nextGaussian() * Math.sqrt(cov), 2) + ",";
            else
                t += Double.NaN + ",";
            // CA-1
            if (random.nextDouble() > 0.001173709)
                t += Util.round(trueChemicalShifts[pos - 1][4] + random.nextGaussian() * Math.sqrt(cav), 2) + ",";
            else
                t += Double.NaN + ",";
            // CB -1 
            if (random.nextDouble() > 0.1709347)
                t += Util.round(trueChemicalShifts[pos - 1][5] + random.nextGaussian() * Math.sqrt(cbv), 2) + ",";
            else
                t += Double.NaN + ",";
            // HA-1
            if (random.nextDouble() > 0.01450494)
                t += Util.round(trueChemicalShifts[pos - 1][6] + random.nextGaussian() * Math.sqrt(hav), 2) + ",";
            else
                t += Double.NaN + ",";
            //HA2-1
            if (random.nextDouble() > 0.01450494)
                t += Util.round(trueChemicalShifts[pos - 1][7] + random.nextGaussian() * Math.sqrt(hav), 2) + ",";
            else
                t += Double.NaN + ",";
            // C
            if (random.nextDouble() > 0.07178364)
                t += Util.round(trueChemicalShifts[pos][3] + random.nextGaussian() * Math.sqrt(cov), 2) + ",";
            else
                t += Double.NaN + ",";
            // CA
            if (random.nextDouble() > 0.002564103)
                t += Util.round(trueChemicalShifts[pos][4] + random.nextGaussian() * Math.sqrt(cav), 2) + ",";
            else
                t += Double.NaN + ",";
            // CB
            if (random.nextDouble() > 0.2310209)
                t += Util.round(trueChemicalShifts[pos][5] + random.nextGaussian() * Math.sqrt(cbv), 2) + ",";
            else
                t += Double.NaN + ",";
            // HA
            if (random.nextDouble() > 0.03232091)
                t += Util.round(trueChemicalShifts[pos][6] + random.nextGaussian() * Math.sqrt(hav), 2) + ",";
            else
                t += Double.NaN + ",";
            // HA2
            if (random.nextDouble() > 0.03232091)
                t += Util.round(trueChemicalShifts[pos][7] + random.nextGaussian() * Math.sqrt(hav), 2) + ",";
            else
                t += Double.NaN + ",";
            // HNv
            t += hv + ","
            // N15v
            +nv + ","
            // COv
            +cov + ","
            // CAv
            +cav + ","
            // CBv
            +cbv + ","
            // HAv
            +hav + ","
            // HA2v
            +hav + ","
            //COv
            +cov + ","
            // CAv
            +cav + ","
            // CBv
            +cbv + ","
            // HAv
            +hav + ","
            // HA2v
            +hav + ",\n";

            s_full += t;

            if (random.nextDouble() < 0.02) {
                simMissingSSat += pos + ",";
                countMissingSS++;
            } else
                s += t;

            // no extra spin system
            if (random.nextDouble() > 0.15)
                continue;
                
            // extra spin system: different from the existing spin systems
             boolean isSimilar = random.nextDouble()  < 0.06906907;
             if (isSimilar == false) {
                                
               Residue r1 = new Residue(sequence.substring(pos - 1, pos));
                r1.simulate();
                Residue r2 = new Residue(sequence.substring(pos, pos + 1));
                r2.simulate();

                t = (pos + 1000) + "," + Util.round(r2.getSimChemicalShift("H"), 2) + "," + Util.round(r2.getSimChemicalShift("N"), 2) + ",";

                // C-1                   
                if (random.nextDouble() > 0.175488)
                    t += Util.round(r1.getSimChemicalShift("C"), 2) + ",";
                else
                    t += Double.NaN + ",";
                 // CA-1                   
                if (random.nextDouble() > 0.3617117)
                    t += Util.round(r1.getSimChemicalShift("CA"), 2) + ",";
                else
                    t += Double.NaN + ",";
                 // CB-1                   
                if (random.nextDouble() > 0.3781907)
                    t += Util.round(r1.getSimChemicalShift("CB"), 2) + ",";
                else
                    t += Double.NaN + ",";
                 // HA-1                   
                if (random.nextDouble() > 0.4354354)
                    t += Util.round(r1.getSimChemicalShift("HA"), 2) + ",";
                else
                    t += Double.NaN + ",";
                 // HA2-1                   
                if (random.nextDouble() > 0.4354354)
                    t += Util.round(r1.getSimChemicalShift("HA2"), 2) + ",";
                else
                    t += Double.NaN + ",";
                 // C                   
                if (random.nextDouble() > 0.6955706)
                    t += Util.round(r2.getSimChemicalShift("C"), 2) + ",";
                else
                    t += Double.NaN + ",";
                 // CA
                if (random.nextDouble() > 0.2337838)
                    t += Util.round(r2.getSimChemicalShift("CA"), 2) + ",";
                else
                    t += Double.NaN + ",";
                 // CB
                if (random.nextDouble() > 0.5110736)
                    t += Util.round(r2.getSimChemicalShift("CB"), 2) + ",";
                else
                    t += Double.NaN + ",";
                 // HA                   
                if (random.nextDouble() > 0.4354354)
                    t += Util.round(r2.getSimChemicalShift("HA"), 2) + ",";
                else
                    t += Double.NaN + ",";
                 // HA2                   
                if (random.nextDouble() > 0.4354354)
                    t += Util.round(r2.getSimChemicalShift("HA2"), 2) + ",";
                else
                    t += Double.NaN + ",";
             }


            // extra spin system: similar to the existing spin systems
            if (isSimilar == true) {
               
                t = (pos + 1000)
                    + ","
                    + Util.round(trueChemicalShifts[pos][1] + random.nextGaussian() * Math.sqrt(hv), 2)
                    + ","
                    + Util.round(trueChemicalShifts[pos][2] + random.nextGaussian() * Math.sqrt(nv), 2)
                    + ",";
            // C-1                   
            if (random.nextDouble() > 0.175488)
                t += Util.round(trueChemicalShifts[pos - 1][3] + random.nextGaussian() * Math.sqrt(cov), 2) + ",";
            else
                t += Double.NaN + ",";
            // CA-1
            if (random.nextDouble() >  0.3617117)
                t += Util.round(trueChemicalShifts[pos - 1][4] + random.nextGaussian() * Math.sqrt(cav), 2) + ",";
            else
                t += Double.NaN + ",";
            // CB -1 
            if (random.nextDouble() > 0.3781907)
                t += Util.round(trueChemicalShifts[pos - 1][5] + random.nextGaussian() * Math.sqrt(cbv), 2) + ",";
            else
                t += Double.NaN + ",";
            // HA-1
            if (random.nextDouble() > 0.4354354)
                t += Util.round(trueChemicalShifts[pos - 1][6] + random.nextGaussian() * Math.sqrt(hav), 2) + ",";
            else
                t += Double.NaN + ",";
            //HA2-1
            if (random.nextDouble() > 0.4354354)
                t += Util.round(trueChemicalShifts[pos - 1][7] + random.nextGaussian() * Math.sqrt(hav), 2) + ",";
            else
                t += Double.NaN + ",";
            // C
            if (random.nextDouble() > 0.6955706)
                t += Util.round(trueChemicalShifts[pos][3] + random.nextGaussian() * Math.sqrt(cov), 2) + ",";
            else
                t += Double.NaN + ",";
            // CA
            if (random.nextDouble() > 0.2337838)
                t += Util.round(trueChemicalShifts[pos][4] + random.nextGaussian() * Math.sqrt(cav), 2) + ",";
            else
                t += Double.NaN + ",";
            // CB
            if (random.nextDouble() > 0.5110736)
                t += Util.round(trueChemicalShifts[pos][5] + random.nextGaussian() * Math.sqrt(cbv), 2) + ",";
            else
                t += Double.NaN + ",";
            // HA
            if (random.nextDouble() > 0.03232091)
                t += Util.round(trueChemicalShifts[pos][6] + random.nextGaussian() * Math.sqrt(hav), 2) + ",";
            else
                t += Double.NaN + ",";
            // HA2
            if (random.nextDouble() > 0.2387387)
                t += Util.round(trueChemicalShifts[pos][7] + random.nextGaussian() * Math.sqrt(hav), 2) + ",";
            else
                t += Double.NaN + ",";
            }
            
                t += hv + "," + nv + "," + cov + "," + cav + "," + cbv + "," + hav + "," + hav + "," + cov + "," + cav + "," + cbv + "," + hav + "," + hav + ",\n";
                s_full += t;
                s += t;
                countExtraSS++;

        }

        // no missing spin systems
        PrintWriter outputStream = null;
        try {
            outputStream = new PrintWriter(new FileOutputStream(outFileMBA_full));
        } catch (FileNotFoundException e) {
            Util.pln("The output file " + outFileMBA_full + " could not be created. The program will now stop");
            System.exit(0);
        }
        // comment and sequence
        outputStream.println(comment);
        outputStream.println(sequence);
        outputStream.println(Integer.MIN_VALUE);
        outputStream.print(s_full);
        outputStream.close();

        // missing spin systems
        try {
            outputStream = new PrintWriter(new FileOutputStream(outFileMBA));
        } catch (FileNotFoundException e) {
            Util.pln("The output file " + outFileMBA + " could not be created. The program will now stop");
            System.exit(0);
        }
        // comment and sequence
        outputStream.println(comment);
        outputStream.println(sequence);
        outputStream.println(simMissingSSat);
        outputStream.print(s);
        outputStream.close();

        Util.pln("There are " + countMissingSS + " missing and " + countExtraSS + " extra spin systems");
    }
    //-----------------------------------------------------------------------------------------
    //---------------------------------------ACCESSORS-----------------------------------------
    //-----------------------------------------------------------------------------------------

    public double getTrueCA(int pos) {
        return trueChemicalShifts[pos][4];
    }

}
