package input;

import util.BitSet;
import util.ChiSquareDistribution;
import util.Util;
import Jama.Matrix;
import assign.Worker;

public class ScoreHandler {

    private Protein protein = Worker.protein;
    private double alpha;

    private double[][] scoreMatch;
    private double[][][] scoreAlign;
    private double[][][] scoreJoint;
    private int[][] dfAlign;
    private int[][] dfMatch;
    private int[][] dfJoint;
    private boolean[][] isValidMatch;
    private boolean[][][] isValidAlign;
    private boolean[][][] isValidJoint;

    private int numberValidPairs;
    private double maxScoreJoint;

    //-----------------------------------------------------------------------
    //---------------------------CONSTRUCTOR---------------------------------
    //-----------------------------------------------------------------------
    public ScoreHandler(boolean insure) {

        alpha = Worker.alpha;
        maxScoreJoint = ChiSquareDistribution.quantile(1 - alpha, protein.computeMaxDf()); //900 for fgf
        int l = protein.getSpinSystems().length;
        int p = protein.getLength();

        scoreMatch = new double[l][l];
        scoreAlign = new double[p][l][l];
        scoreJoint = new double[p][l][l];
        dfAlign = new int[l][l];
        dfMatch = new int[l][l];
        dfJoint = new int[l][l];
        isValidMatch = new boolean[l][l];
        isValidAlign = new boolean[p][l][l];
        isValidJoint = new boolean[p][l][l];
        numberValidPairs = 0;

        for (int i = 0; i < l; i++) {
            for (int j = 0; j < l; j++) {

                SpinSystem within = protein.getSpinSystems()[i];
                SpinSystem seq = protein.getSpinSystems()[j];
                boolean order = orderHA(within.getHAw(), within.getHA2w(), seq.getHAs(), seq.getHA2s(), within.getHAwv(), within.getHA2wv(), seq.getHAsv(), seq.getHA2sv());

                scoreMatch[i][j] = computeScoreMatch(within, seq, order);
                dfMatch[i][j] = computeDfMatch(within, seq, order);
                isValidMatch[i][j] = isValid(scoreMatch[i][j], dfMatch[i][j]);

                dfAlign[i][j] = computeDfAlign(within, seq, order);
                dfJoint[i][j] = dfAlign[i][j] + dfMatch[i][j];
                for (int pos = 0; pos < p; pos++) {

                    if (!isValidMatch[i][j]) {
                        scoreAlign[pos][i][j] = -1;
                        isValidAlign[pos][i][j] = false;

                        scoreJoint[pos][i][j] = -1;
                        isValidJoint[pos][i][j] = false;

                        continue;
                    } else {
                        scoreAlign[pos][i][j] = computeScoreAlign(within, seq, order, pos);
                        isValidAlign[pos][i][j] = isValid(scoreAlign[pos][i][j], dfAlign[i][j]);

                        scoreJoint[pos][i][j] = scoreAlign[pos][i][j] + scoreMatch[i][j];
                        isValidJoint[pos][i][j] = isValid(scoreJoint[pos][i][j], dfJoint[i][j]);
                    }

                    if (isValidMatch[i][j] && isValidAlign[pos][i][j] && isValidJoint[pos][i][j])
                        numberValidPairs++;
                }

            }
        }
        
        if (insure)
            insureUniqueMatch();
    }

    //-----------------------------------------------------------------------
    //------------------------------METHODS----------------------------------
    //-----------------------------------------------------------------------
    // -------------is the score valid-------------------------------------------
    public boolean isValid(double score, int df) {

        if (df < 0 || score < 0)
            return false;

        if (df == 0 && score == 0)
            return true;

        if (df == 0 && score > 0)
            return false;

        return score < ChiSquareDistribution.quantile(1 - alpha, df);
    }

    // -------------df of alignment-------------------------------------------
    public int computeDfAlign(SpinSystem within, SpinSystem seq, boolean order) {

        int df = 0;
        if (!Double.isNaN(within.getHw()))
            df += 1;

        if (!Double.isNaN(within.getNw()))
            df += 1;

        if (!Double.isNaN(within.getCw()) || !Double.isNaN(seq.getCs()))
            df += 1;

        if (!Double.isNaN(within.getCAw()) || !Double.isNaN(seq.getCAs()))
            df += 1;

        if (!Double.isNaN(within.getCBw()) || !Double.isNaN(seq.getCBs()))
            df += 1;

        if (!Double.isNaN(within.getHAw()) || !Double.isNaN(seq.getHAs())) {

            if (order) {
                df += 1;
                if (!Double.isNaN(within.getHA2w()) || !Double.isNaN(seq.getHA2s()))
                    df += 1;
            } else {
                if (!Double.isNaN(within.getHAw()) || !Double.isNaN(seq.getHA2s()))
                    df += 1;
                if (!Double.isNaN(within.getHA2w()) || !Double.isNaN(seq.getHAs()))
                    df += 1;
            }
        }
        return df;
    }

    // -------------df of match-------------------------------------------
    private int computeDfMatch(SpinSystem within, SpinSystem seq, boolean order) {

        int df = 0;
        if (!Double.isNaN(within.getCw()) && !Double.isNaN(seq.getCs()))
            df += 1;

        if (!Double.isNaN(within.getCAw()) && !Double.isNaN(seq.getCAs()))
            df += 1;

        if (!Double.isNaN(within.getCBw()) && !Double.isNaN(seq.getCBs()))
            df += 1;

        if (!Double.isNaN(within.getHAw()) && !Double.isNaN(seq.getHAs())) {

            if (order) {
                df += 1;
                if (!Double.isNaN(within.getHA2w()) && !Double.isNaN(seq.getHA2s()))
                    df += 1;
            } else {
                if (!Double.isNaN(within.getHAw()) && !Double.isNaN(seq.getHA2s()))
                    df += 1;
                if (!Double.isNaN(within.getHA2w()) && !Double.isNaN(seq.getHAs()))
                    df += 1;
            }
        }
        return df;
    }

    // --------------score of alignment------------------------------------------
    public double computeScoreAlign(SpinSystem within, SpinSystem seq, boolean order, int pos) throws Error {

        // alignment before an unassignable position
        if (protein.getAssignablePositions().test(pos + 1) == false && !seq.isEmpty())
            return -1;

        // alignment to an unassignable position
        if (protein.getAssignablePositions().test(pos) == false && !within.isEmpty())
            return -1;

        // alignment to the last position
        if (pos == protein.getLastPos() && !seq.isEmpty())
            return -1;

        // restrictions for Gly
        Residue residue = protein.getResidueAt(pos);
        if (residue.isGlycine() && !within.canBeGly())
            return -1;
        if (!residue.isGlycine() && !within.canBeNonGly())
            return -1;
        if (residue.isGlycine() && !seq.canFollowGly())
            return -1;
        if (!residue.isGlycine() && !seq.canFollowNonGly())
            return -1;

        // compute the score
        //H, HA, HA2, HA3, C, CA, CB, N
        Matrix mean = residue.getMean().copy();
        Matrix var = residue.getVar().copy();
        String[] dimNames = residue.getDimNames();
        BitSet dimUsed = new BitSet(0, var.getColumnDimension());

        for (int i = 0; i < dimNames.length; i++) {

            if (dimNames[i] == "H" && !Double.isNaN(within.getHw())) {
                mean.set(i, 0, updateMean(within.getHw(), Double.NaN, mean.get(i, 0)));
                var.set(i, i, updateVar(within.getHw(), Double.NaN, within.getHwv(), Double.NaN, var.get(i, i)));
                dimUsed.set(i);
            }

            if (dimNames[i] == "N" && !Double.isNaN(within.getNw())) {
                mean.set(i, 0, updateMean(within.getNw(), Double.NaN, mean.get(i, 0)));
                var.set(i, i, updateVar(within.getNw(), Double.NaN, within.getNwv(), Double.NaN, var.get(i, i)));
                dimUsed.set(i);
            }

            if (dimNames[i] == "C" && (!Double.isNaN(within.getCw()) || !Double.isNaN(seq.getCs()))) {
                mean.set(i, 0, updateMean(within.getCw(), seq.getCs(), mean.get(i, 0)));
                var.set(i, i, updateVar(within.getCw(), seq.getCs(), within.getCwv(), seq.getCsv(), var.get(i, i)));
                dimUsed.set(i);
            }

            if (dimNames[i] == "CA" && (!Double.isNaN(within.getCAw()) || !Double.isNaN(seq.getCAs()))) {
                mean.set(i, 0, updateMean(within.getCAw(), seq.getCAs(), mean.get(i, 0)));
                var.set(i, i, updateVar(within.getCAw(), seq.getCAs(), within.getCAwv(), seq.getCAsv(), var.get(i, i)));
                dimUsed.set(i);
            }

            if (dimNames[i] == "CB" && (!Double.isNaN(within.getCBw()) || !Double.isNaN(seq.getCBs()))) {
                mean.set(i, 0, updateMean(within.getCBw(), seq.getCBs(), mean.get(i, 0)));
                var.set(i, i, updateVar(within.getCBw(), seq.getCBs(), within.getCBwv(), seq.getCBsv(), var.get(i, i)));
                dimUsed.set(i);
            }

            if (dimNames[i] == "HA" && (!Double.isNaN(within.getHAw()) || !Double.isNaN(seq.getHAs()))) {
                mean.set(i, 0, updateMean(within.getHAw(), seq.getHAs(), mean.get(i, 0)));
                var.set(i, i, updateVar(within.getHAw(), seq.getHAs(), within.getHAwv(), seq.getHAsv(), var.get(i, i)));
                dimUsed.set(i);
            }

            if (dimNames[i] == "HA2" && order && (!Double.isNaN(within.getHAw()) || !Double.isNaN(seq.getHAs()))) {
                mean.set(i, 0, updateMean(within.getHAw(), seq.getHAs(), mean.get(i, 0)));
                var.set(i, i, updateVar(within.getHAw(), seq.getHAs(), within.getHAwv(), seq.getHAsv(), var.get(i, i)));
                dimUsed.set(i);
            }

            if (dimNames[i] == "HA2" && !order && (!Double.isNaN(within.getHAw()) || !Double.isNaN(seq.getHA2s()))) {
                mean.set(i, 0, updateMean(within.getHAw(), seq.getHA2s(), mean.get(i, 0)));
                var.set(i, i, updateVar(within.getHAw(), seq.getHA2s(), within.getHAwv(), seq.getHA2sv(), var.get(i, i)));
                dimUsed.set(i);
            }

            if (dimNames[i] == "HA3" && order && (!Double.isNaN(within.getHA2w()) || !Double.isNaN(seq.getHA2s()))) {
                mean.set(i, 0, updateMean(within.getHA2w(), seq.getHA2s(), mean.get(i, 0)));
                var.set(i, i, updateVar(within.getHA2w(), seq.getHA2s(), within.getHA2wv(), seq.getHA2sv(), var.get(i, i)));
                dimUsed.set(i);
            }

            if (dimNames[i] == "HA3" && !order && (!Double.isNaN(within.getHA2w()) || !Double.isNaN(seq.getHAs()))) {
                mean.set(i, 0, updateMean(within.getHA2w(), seq.getHAs(), mean.get(i, 0)));
                var.set(i, i, updateVar(within.getHA2w(), seq.getHAs(), within.getHA2wv(), seq.getHAsv(), var.get(i, i)));
                dimUsed.set(i);
            }

        }

        if (dimUsed.cardinality() == 0)
            return 0;

        mean = Util.subset(mean, dimUsed, null);
        var = Util.subset(var, dimUsed, dimUsed);
        if (var.rank() < var.getColumnDimension())
            throw new Error("Singular variance matrix");
        return mean.transpose().times(var.inverse()).times(mean).get(0, 0);

        //-----
    }

    // --------------score of match----------------------------------------------
    private double computeScoreMatch(SpinSystem within, SpinSystem seq, boolean order) {

        // match of a spin system with itself is impossible
        if (!seq.isEmpty() && within.getReferenceId() == seq.getReferenceId())
            return -1000;

        double score = 0;

        if (!Double.isNaN(within.getCw()) && !Double.isNaN(seq.getCs()))
            score += computeScore(within.getCw(), seq.getCs(), within.getCwv(), seq.getCsv());

        if (!Double.isNaN(within.getCAw()) && !Double.isNaN(seq.getCAs()))
            score += computeScore(within.getCAw(), seq.getCAs(), within.getCAwv(), seq.getCAsv());

        if (!Double.isNaN(within.getCBw()) && !Double.isNaN(seq.getCBs()))
            score += computeScore(within.getCBw(), seq.getCBs(), within.getCBwv(), seq.getCBsv());

        if (!Double.isNaN(within.getHAw()) && !Double.isNaN(seq.getHAs())) {
            if (order) {
                score += computeScore(within.getHAw(), seq.getHAs(), within.getHAwv(), seq.getHAsv());
                score += computeScore(within.getHA2w(), seq.getHA2s(), within.getHA2wv(), seq.getHA2sv());
            } else {
                score += computeScore(within.getHAw(), seq.getHA2s(), within.getHAwv(), seq.getHA2sv());
                score += computeScore(within.getHA2w(), seq.getHAs(), within.getHA2wv(), seq.getHAsv());
            }
        }

        return score;

    }

    // -------------------score of match in one dimension-------------------------------
    private double computeScore(double x1, double x2, double x1v, double x2v) {

        if (!Double.isNaN(x1) && !Double.isNaN(x2)) {
            double x_sub = x1 - x2;
            return (x_sub * x_sub) / (x1v + x2v);
        } else
            return 0;
    }

    // ----compute mean chemical shift in one dimension for alignment with covariance----
    private double updateMean(double x1, double x2, double theta) {

        if (Double.isNaN(x1) && !Double.isNaN(x2)) {
            return x2 - theta;

        } else if (!Double.isNaN(x1) && Double.isNaN(x2)) {
            return x1 - theta;

        } else if (!Double.isNaN(x1) && !Double.isNaN(x2)) {
            return (x1 + x2) / 2 - theta;

        } else
            return Double.MAX_VALUE;

    }

    // ---update the posterior variance in one dimension for alignment with covariance---
    private double updateVar(double x1, double x2, double x1v, double x2v, double sigma) {

        if (Double.isNaN(x1) && !Double.isNaN(x2)) {
            return sigma + x2v;

        } else if (!Double.isNaN(x1) && Double.isNaN(x2)) {
            return sigma + x1v;

        } else if (!Double.isNaN(x1) && !Double.isNaN(x2)) {
            return sigma + x1v + x1v / 4 + x2v / 4;
        }
        return -1;

    }

    // -----------determine the order of HA and HA2--------------------
    private boolean orderHA(double x11, double x12, double x21, double x22, double x11v, double x12v, double x21v, double x22v) {

        if (!Double.isNaN(x12) || !Double.isNaN(x22)) {
            double score11 = computeScore(x11, x21, x11v, x21v);
            double score22 = computeScore(x12, x22, x12v, x22v);
            double score12 = computeScore(x11, x22, x11v, x22v);
            double score21 = computeScore(x12, x21, x12v, x21v);

            if (score11 + score22 <= score12 + score21)
                return true;
            else
                return false;
        } else
            return true;
    }

    private void insureUniqueMatch() {

        BitSet uniqueMatch = new BitSet(0, protein.getSpinSystems().length);
        for (int i = 0; i < protein.getSpinSystems().length; i++) {
            if (isUniquelyFollowed(i))
                uniqueMatch.set(i);
        }

        //set unique
        for (int i = 0; i < protein.getSpinSystems().length; i++)
            if (uniqueMatch.test(i))
                for (int j = 0; j < protein.getSpinSystems().length; j++) {
                    isValidMatch[i][j] = false;
                    isValidMatch[j][i + 1] = false;
                    isValidMatch[i][i + 1] = true;
                }
    }

    //-----------------------------------------------------------------------
    //-----------------------COMPUTATION OF STATISTICS-----------------------
    //-----------------------------------------------------------------------
    public String toStringStats() {

        double meanNumberFollowingSpinSystems = 0;
        for (int i = 0; i < protein.getSpinSystems().length - 1; i++) {
            meanNumberFollowingSpinSystems += computeNumberFollowingSpinSystems(i);
        }
        meanNumberFollowingSpinSystems = meanNumberFollowingSpinSystems / (protein.getSpinSystems().length - 1);

        int numberUniquelyFollowedSpinSystems = 0;
        for (int i = 0; i < protein.getSpinSystems().length - 1; i++) {
            if (isUniquelyFollowed(i))
                numberUniquelyFollowedSpinSystems++;
        }

        String s =
            "\n\nDESCRIPTION OF THE SCORES:\nNumber of valid pairs = "
                + numberValidPairs
                + "\nMean number of following spin systems: "
                + Util.round(meanNumberFollowingSpinSystems, 2)
                + ";\nNumber of uniquely followed spin systems: "
                + numberUniquelyFollowedSpinSystems
                + " (Prop out of number of non-missing SS: "
                + Util.round(((double) numberUniquelyFollowedSpinSystems * 100) / (protein.getSpinSystems().length - 1), 2)
                + "%).\n";

        return s;

    }

    private int computeNumberFollowingSpinSystems(int spinSystemId) {
        int result = 0;
        for (int i = 0; i < protein.getSpinSystems().length - 1; i++)
            if (isValidMatch[spinSystemId][i])
                result++;
        return result;
    }

    private int computeNumberPreceedingSpinSystems(int spinSystemId) {
        int result = 0;
        for (int i = 0; i < protein.getSpinSystems().length - 1; i++)
            if (isValidMatch[i][spinSystemId])
                result++;
        return result;
    }

    private boolean isUniquelyFollowed(int spinSystemId) {

        if (computeNumberFollowingSpinSystems(spinSystemId) != 1)
            return false;

        int i;
        for (i = 0; i < protein.getSpinSystems().length - 1; i++)
            if (isValidMatch[spinSystemId][i] && i != spinSystemId)
                break;

        if (computeNumberPreceedingSpinSystems(i) == 1)
            return true;
        else
            return false;
    }

    //-----------------------------------------------------------------------
    //-----------------------------ACCESSORS---------------------------------
    //-----------------------------------------------------------------------
    public double getScoreAlign(int pos, SpinSystem within, SpinSystem seq) {
        return scoreAlign[pos][within.getId()][seq.getId()];
    }

    public double getScoreMatch(SpinSystem within, SpinSystem seq) {
        return scoreMatch[within.getId()][seq.getId()];
    }

    public double getScoreJoint(int pos, SpinSystem within, SpinSystem seq) {
        return scoreJoint[pos][within.getId()][seq.getId()];
    }

    public int getDfAlign(SpinSystem within, SpinSystem seq) {
        return dfAlign[within.getId()][seq.getId()];
    }
    public int getDfMatch(SpinSystem within, SpinSystem seq) {
        return dfMatch[within.getId()][seq.getId()];
    }

    public int getDfJoint(SpinSystem within, SpinSystem seq) {
        return dfJoint[within.getId()][seq.getId()];
    }

    public boolean isValidAlign(int pos, SpinSystem within, SpinSystem seq) {
        return isValidAlign[pos][within.getId()][seq.getId()];
    }

    public boolean isValidMatch(SpinSystem within, SpinSystem seq) {
        return isValidMatch[within.getId()][seq.getId()];
    }

    public boolean isValidJoint(int pos, SpinSystem within, SpinSystem seq) {
        return isValidJoint[pos][within.getId()][seq.getId()];
    }

    public double getMaxScoreJoint() {
        return maxScoreJoint;
    }

}
