package output;

import input.Protein;

import java.util.Arrays;
import java.util.HashSet;
import java.util.Iterator;
import java.util.Set;

import util.BitSet;
import util.ChiSquareDistribution;
import util.Util;
import util.Util.Pair;
import assign.Strand;
import assign.Worker;

public class MappingsFound {

    // assign
    private Set strandSet; // a Set ensures we don't have duplicates
    private boolean dirty; // true if strands[] and penalizedScores[] are out of date
    private Strand[] strands;
    private double[] penalizedScores;
    private double maxPenalizedScore;

    // analyze
    private BitSet[] summary;

    //context
    private static Protein protein = Worker.protein;

    //-----------------------------------------------------------------------------------
    //--------------------------------CONSTRUCTORS---------------------------------------
    //-----------------------------------------------------------------------------------
    //------------------------------initialize to empty----------------------------------
    public MappingsFound() {
        // assign
        strandSet = new HashSet();
        dirty = true;
        strands = null;
        penalizedScores = null;
        maxPenalizedScore = Worker.scoreHandler.getMaxScoreJoint();

        // analyze
        summary = new BitSet[protein.getLength()];
        for (int pos = 0; pos < protein.getLength(); pos++)
            summary[pos] = new BitSet(0, protein.getSpinSystems().length);

    }

    //-----------------------------------------------------------------------------------
    //-------------------------------------METHODS---------------------------------------
    //-----------------------------------------------------------------------------------
    //-----------------------------------add strands-------------------------------------
    public synchronized void addStrands(Strand[] newStrands) {

        if (newStrands == null)
            return;

        //compute the incoming penalized joint scores and add the strands to the set
        for (int i = 0; i < newStrands.length; i++) {

            double score = newStrands[i].getScoreJoint();
            int penaltyDf = protein.computeMaxDf() - newStrands[i].getDfJoint();
            double penalty = (penaltyDf > 0) ? ChiSquareDistribution.quantile(1 - Worker.alpha, penaltyDf) : 0;

            strandSet.add(new Pair(newStrands[i], score + penalty));
        }

        //find minimum penalized score
        double minPenalizedScoreFound = Double.MAX_VALUE;
        Iterator it = strandSet.iterator();
        while (it.hasNext()) {
            Pair pair = (Pair) it.next();
            if (pair.getMetric() < minPenalizedScoreFound)
                minPenalizedScoreFound = pair.getMetric();
        }

        //maxScoreJoint
        maxPenalizedScore = minPenalizedScoreFound + Worker.deltaScore;

        //remover the strands with penalized score > minPenalizedScore + deltaScore
        it = strandSet.iterator();
        while (it.hasNext()) {
            Pair pair = (Pair) it.next();
            if (pair.getMetric() > maxPenalizedScore)
                it.remove();
        }

        Util.pln("\n*************** Now  have " + strandSet.size() + " completed mappings ***************");
        dirty = true;
        return;
    }

    public String toStringByPosition() {
        sortStrands();

        String s = "";
        if (strands == null)
            return "\nNo completed mappings found\n";

        s += "\n\n\nMAPPINGS FOUND: SUMMARY\nThere are " + strands.length + " solution(s).\n";

        // compute summaries
        computeSummary();
        for (int pos = 0; pos < protein.getLength(); pos++) {
            s += protein.getResidueAt(pos).getShortName() + pos + "  *  [";

            for (int i = 0; i < protein.getSpinSystems().length; i++) {
                if (summary[pos].test(i) == true)
                    s += protein.getSpinSystems()[i].getReferenceId() + " ";
            }
            s += "]\n";
        }
        return s;
    }

    public String toStringByMapping() {
        sortStrands();

        if (strands == null)
            return "";

        String s = "\n\n\nMAPPINGS FOUND: DETAILS\n\n";

        for (int i = 0; i < strands.length; i++) {

            if (strands[i].isReference())
                s += "########## EXPECTED ############\n";

            s += "Assignment:     "
                + strands[i].toString()
                + "\nScore:          "
                + strands[i].getScoreJoint()
                + "\nPenalizedScore: "
                + penalizedScores[i]
                + "\nDf:             "
                + strands[i].getDfJoint()
                + "\n\n";
        }
        return s;
    }
    
    public String toStringByResonances() {
        
        String s = "\n\n\nMAPPINGS FOUND: RESONANCES\n\n";
        return s;
    }

    public void stats() {
        Util.pln(toStringByPosition());
        Util.pln("\n");
        Util.pln(toStringByMapping());
    }

    private double computeOneStrandPredMeanCA(int j, int pos) {
        double result = 0;
        int count = 0;

        result += Double.isNaN(strands[j].getSpinSystems()[pos].getCAw()) ? 0 : strands[j].getSpinSystems()[pos].getCAw();
        if (pos != protein.getLastPos())
            result += Double.isNaN(strands[j].getSpinSystems()[pos + 1].getCAs()) ? 0 : strands[j].getSpinSystems()[pos + 1].getCAs();

        count += Double.isNaN(strands[j].getSpinSystems()[pos].getCAw()) ? 0 : 1;
        if (pos != protein.getLastPos())
            count += Double.isNaN(strands[j].getSpinSystems()[pos + 1].getCAs()) ? 0 : 1;

        result = (count == 0) ? Double.NaN : (result / count);
        return result;
    }

    private double computeOneStrandPredSdCA(int j, int pos) {
        double result = 0;
        int count = 0;

        result += Double.isNaN(strands[j].getSpinSystems()[pos].getCAw()) ? 0 : strands[j].getSpinSystems()[pos].getCAwv();
        if (pos != protein.getLastPos())
            result += Double.isNaN(strands[j].getSpinSystems()[pos + 1].getCAs()) ? 0 : strands[j].getSpinSystems()[pos + 1].getCAsv();

        count += Double.isNaN(strands[j].getSpinSystems()[pos].getCAw()) ? 0 : 1;
        if (pos != protein.getLastPos())
            count += Double.isNaN(strands[j].getSpinSystems()[pos + 1].getCAs()) ? 0 : 1;

        result = (count == 0) ? Double.NaN : (result / (count * count));
        return result;
    }

    private void computeSummary() {

        for (int pos = 0; pos < protein.getLength(); pos++)
            for (int j = 0; j < strands.length; j++)
                summary[pos].set(strands[j].getSpinSystemIdAt(pos));

    }

    private void sortStrands() {
        //sort the strand by increasing penalized score, if necessary

        if (!dirty)
            return;

        Pair[] array = (Pair[]) strandSet.toArray(new Pair[0]);
        Arrays.sort(array);

        strands = new Strand[array.length];
        penalizedScores = new double[array.length];
        for (int i = 0; i < array.length; i++) {
            strands[i] = (Strand) array[i].getObject();
            penalizedScores[i] = array[i].getMetric();
        }

        dirty = false;
    }
    //-----------------------------------------------------------------------------------
    //------------------------------------ACCESSORS--------------------------------------
    //-----------------------------------------------------------------------------------
    public Strand[] getStrands() {
        sortStrands();
        return strands;
    }

    public double getMaxPenalizedScore() {
        return maxPenalizedScore;
    }
}
