package parallel;

import util.Util;

import java.io.Serializable;

import EDU.oswego.cs.dl.util.concurrent.FJTask;

/**
 * A specialized variant of FJTask designed to be capable of being moved from
 * one process to another.  For this to work, the work done by the task must
 * not depend on seeing the results of any subtasks it launches.  One way to
 * do this is to have tasks report any results to some centralized location.
 * 
 * To help with garbage collection, subclasses that carry large amounts of data
 * should override cancel() with code that nulls out any data before calling
 * the superclass's version.
 * 
 * It's a very bad idea to try to run the same MovableTask object twice, since
 * they will typically destroy references to the data they use once the data
 * are no longer needed.  This is because MovableTask objects can live for a
 * long time, and deleting references to unneeded data helps the garbage
 * collector.
 * 
 * @author kulinipm
 */
public abstract class MovableTask extends FJTask implements Serializable {

	// Create a task that will produce the same result as this task, but that
	// is more suitable for sending across the network.  A typical implementation
	// might return a task that recomputes the data stored in this task before
	// executing a copy of the original task.  This permits recomputation of
	// data on the remote system to reduce network overhead.
	public abstract MovableTask createRemoteEquivalent();
	
	// Subclasses should implement this instead of run().
	protected abstract void runImpl();
	
	// Gracefully handle OOM conditions by trying to free up some more memory
	// before retrying the task or, as a last result, scaling back the amount
	// of parallelism.
	public final void run() {
		try {
			runImpl();
			if (Thread.interrupted())
				throw new MovableTaskRunner.TerminateException(true);
		}
		catch (OutOfMemoryError e) {
			// This system appears to have insufficient resources.  Offload the
			// task and kill off the current task runner.
			Util.pln("Caught OutOfMemoryError; offloading the offending task");
			offload();
			throw new MovableTaskRunner.TerminateException(false);
		}
	}
	
	public void offload() {
		Util.debugln("Offloading a " + getClass().getName());
		MovableTask copy = createRemoteEquivalent();
		cancel();
		Parallelizer.obj.enqueueTask(copy);
	}
}
