package parallel;

import util.Util;

import EDU.oswego.cs.dl.util.concurrent.FJTaskRunnerGroup;

import java.util.Random;

/**
 * A specialized variant of FJTaskRunnerGroup that allows tasks to be stolen
 * by an external entity.  In a distributed system, this allows computers
 * who need more tasks to take them from computers with a backlog of tasks.
 * 
 * This class only works with tasks derived from MovableTask, since MovableTask
 * defines additional interfaces needed when stealing them from the group.
 * 
 * @author kulinipm
 */
public class MovableTaskRunnerGroup extends FJTaskRunnerGroup {
	
	private final Random victimRNG;
	private int numAlive;
	
	// for stats
	private int numRemoteScans;
	private int numRemoteSteals;

	public MovableTaskRunnerGroup(int groupSize) {
		super(groupSize);
		victimRNG = new Random(System.identityHashCode(this));
		numAlive = groupSize;
		numRemoteScans = 0;
		numRemoteSteals = 0;
	}
	
	public void stats() {
		super.stats();
		System.out.println("Remote scans: " + numRemoteScans);
		System.out.println("Remote steals: " + numRemoteSteals);
	}
	
	public synchronized void threadDied() {
		Util.debugln("One of the task runners died; ");
		numAlive--;
		
		if (numAlive == 0) {
			Util.pln("!!!  ALL WORKER THREADS HAVE DIED PREMATURELY  !!!");
			Util.pln("!!! UNABLE TO CONTINUE WORKING ON THE SOLUTION !!!");
			Parallelizer.obj.abort();
		}
	}
	
	// makes sure all threads have terminated before returning
	public void interruptAll() {
		Util.debugln("Terminating all worker threads");
		super.interruptAll();
		for (int i = 0; i < threads.length; i++) {
			if (threads[i] != Thread.currentThread()) {
				try {
					threads[i].join();
				}
				catch (InterruptedException e) {
					// ignore
				}
			}
		}
	}
	
	//------steal a task for remote execution, if possible------
	// implementation based on that of FJTaskRunner.scan(FJTask)
	public MovableTask steal() {
		MovableTask task = null;
		
		int idx = victimRNG.nextInt(threads.length);
		for (int i = 0; i < threads.length; i++) {
			MovableTaskRunner thread = (MovableTaskRunner) threads[idx];
			if (++idx >= threads.length)
				idx = 0;
			numRemoteScans++;
			task = thread.takeProxy();
			if (task != null) {
				numRemoteSteals++;
				break;
			}
		}
		
		if (task == null)
			return null;
		
		MovableTask result = task.createRemoteEquivalent();
		task.cancel();
		return result;
	}
	
	//-----override thread creation to use MovableTaskRunner threads------
	protected void initializeThreads() {
		for (int i = 0; i < threads.length; i++)
			threads[i] = new MovableTaskRunner(this);
	}
}
