package parallel;

import assign.Worker.Argument;
import util.Util;

import EDU.oswego.cs.dl.util.concurrent.Channel;
import EDU.oswego.cs.dl.util.concurrent.LinkedQueue;

/**
 * Base class for parallelization support.  Subclasses have the responsibility
 * of deciding how to distribute the workload across available resources.
 * 
 * @author kulinipm
 */
public abstract class Parallelizer {

	// The parallelizer that the program should use.
	public static Parallelizer obj;
	
	protected Channel tasks = new LinkedQueue();
	
	// ---------------------- initialization ---------------------------
	public static boolean initialize(Argument arg) {
		
		// Decide which parallelizer to use.
		if (arg.mode.equals("local"))
			obj = new ThreadedParallelizer(arg);
		else if (arg.mode.equals("server")) {
			try {
				obj = new RmiServerParallelizer(arg);
			}
			catch (Exception e) {
				Util.pln("Couldn't start server: " + e.getMessage());
				e.printStackTrace();
			}
			return false;
		}
		else if (arg.mode.equals("client")) {
			try {
				obj = new RmiClientParallelizer(arg);
			}
			catch (Exception e) {
				Util.pln("Couldn't start client: " + e.getMessage());
				e.printStackTrace();
				return false;
			}
		}
		else {
			Util.pln("WARN: mode \"" + arg.mode + "\" is unrecognized; assuming local");
			obj = new ThreadedParallelizer(arg);
		}
		return true;
	}
	
	//-----------------------------------------------------------------------
	//-------------------------SUBCLASS INTERFACE----------------------------
	//-----------------------------------------------------------------------
	
	// Begin the computation.
	public abstract void start();

	// Schedule another task for parallelization.
	public void enqueueTask(MovableTask task) {
		try {
			tasks.put(task);
		}
		catch (InterruptedException e) {
			Thread.currentThread().interrupt();
		}
	}

	// Get the next task available for parallelization.
	public MovableTask dequeueTask() {
		try {
			return (MovableTask) tasks.poll(0);
		}
		catch (InterruptedException e) {
			Thread.currentThread().interrupt();
			return null;
		}
	}
	
	// Handles the case where the program must abort computation, probably
	// because if unsufficient resources.
	public void abort() {
		Util.pln("!!! ABORTING COMPUTATION !!!");
	}
	
	// Called when all the work is done, to release non-GC'd resources.
	public void finished() {}
}
