package parallel;

import assign.SearchByFixedPosTask;
import assign.Strand;
import assign.Worker;
import assign.Worker.Argument;
import util.Util;

import java.net.SocketTimeoutException;
import java.rmi.Naming;
import java.rmi.RemoteException;
import java.rmi.server.UnicastRemoteObject;
import java.util.Random;

/**
 * Parallelizer that distributes tasks to one or more RmiServers to do the
 * actual processing.  To simplify things, no actual processing (aside from
 * merging results) takes place on the client.  If you want the client machine
 * to also execute tasks, just run an instance of the server on the same
 * machine too.
 * 
 * @author kulinipm
 */
public class RmiClientParallelizer extends Parallelizer implements RmiClient {
	
	private Connection[] servers;
	private int numRunning;
	private int numBusy;
	
	private Argument arg;
	
	private final Random victimRNG;
	
	//-----------------------------------------------------------------------------------
	//-----------------------------CONSTRUCTORS------------------------------------------
	//-----------------------------------------------------------------------------------
	
	public RmiClientParallelizer(Argument arg) throws RemoteException {
		
		//------------------make this client visible via RMI----------------------------
		UnicastRemoteObject.exportObject(this);
		
		//------------establish connections to the servers-------------------
		String[] serverNames = arg.servers.split(",");
		numRunning = 0;
		numBusy = 0;
		
		Util.pln("\n");
		
		servers = new Connection[serverNames.length];
		for (int i = 0; i < serverNames.length; i++)
			servers[i] = new Connection(serverNames[i], i);

		//---------------other setup---------------------------
		this.arg = arg;
		Worker.initialize(arg);		// needed for MappingsFound() constructor
		
		victimRNG = new Random(System.identityHashCode(this));
	}
		
	//-----------------------------------------------------------------------------------
	//--------------------------SUPERCLASS METHODS---------------------------------------
	//-----------------------------------------------------------------------------------
	
	public void start() {
		
		enqueueTask((new SearchByFixedPosTask()).createRemoteEquivalent());
		
		for (int i = 0; i < servers.length; i++)
			servers[i].start();
		
		(new Reaper()).start();
		
		waitUntilServersAreDone();
		
		for (int i = 0; i < servers.length; i++)
			servers[i].close(true);

		if (allServersDied())
			Util.pln("\n\n!!! ALL SERVERS DIED PREMATURELY; RESULT IS INCOMPLETE !!!\n\n");
	}
	
	public void finished() {
		for (int i = 0; i < servers.length; i++)
			servers[i].close(true);
	}
	
	//-----------------------------------------------------------------------------------
	//---------------------------CLIENT INTERFACE----------------------------------------
	//-----------------------------------------------------------------------------------	

	public void enqueueTask(int i, MovableTask task) {
		Util.debugln(servers[i].getName() + " is enqueueing a task");
		servers[i].setAlive();
		enqueueTask(task);
	}

	public synchronized MovableTask fetchTask(int i) {
		Util.debugln(servers[i].getName() + " is asking for more work");
		servers[i].setAlive();
		
		MovableTask task = dequeueTask();
		
		if (task != null)
			notify();
		else {
			// Try to take a task from a busy server and reassign it.
			
			int idx = victimRNG.nextInt(servers.length);
			for (int j = 0; j < servers.length; j++) {
				Connection server = servers[idx];
				if (++idx >= servers.length)
					idx = 0;
				if (server == servers[i])
					continue;
				task = server.stealTask();
				if (task != null)
					break;
			}
		}
		
		if (task != null)
			Util.debugln("     ***** Sending " + servers[i].getName() + " another task");
		else
			Util.debugln("     No work is available for " + servers[i].getName());
		
		servers[i].setCurrentTask(task);
		return task;
	}
	
	public void addStrands(int i, Strand[] strands) {
		Util.debugln("Receiving " + strands.length + " strands from " + servers[i].getName());
		servers[i].setAlive();
		
		Worker.mappingsFound.addStrands(strands);
	}
	
	public void ping(int i) {
		servers[i].setAlive();
	}
	
	public void close(int i) {
		Util.debugln(servers[i].getName() + " is closing its connection to the client.");
		servers[i].remoteClose();
	}
	
	//-----------------------------------------------------------------------------------
	//----------------------------MISCELLANEOUS------------------------------------------
	//-----------------------------------------------------------------------------------
	
	private synchronized void waitUntilServersAreDone() {
		try {
			while (numRunning > 0 && (numBusy > 0 || tasks.peek() != null))
				wait();
		}
		catch (InterruptedException e) {
			// ignore
		}
	}
	
	private boolean allServersDied() {
		for (int i = 0; i < servers.length; i++) {
			if (servers[i].getState() == State.CLOSED)
				return false;
		}
		return true;
	}
	
	// fancy enumeration of possible connection states
	private static class State {
		private static final State UNAVAILABLE = new State(false);	// couldn't connect
		private static final State READY = new State(true);			// waiting for something to do
		private static final State BUSY = new State(true);			// busy computing something
		private static final State DIED = new State(false);			// died mysteriously
		private static final State CLOSED = new State(false);		// closed cleanly
		
		private final boolean isUp;
		
		private State(boolean isUp) {
			this.isUp = isUp;
		}
		
		public boolean isUp() {
			return isUp;
		}
		
		public boolean isDown() {
			return !isUp;
		}
	}
	
	private class Connection {
		
		private static final int MAX_STRIKES = 3;
		
		private String name;
		private int id;
		private RmiServer server;
		private State state;
		private MovableTask currentTask;
		private boolean isAlive;
		
		public Connection(String name, int id) {
			this.name = name;
			this.id = id;
			state = State.UNAVAILABLE;
			currentTask = null;
			
			try {
				String resource = "//" + name + "/RmiServer";
				server = (RmiServer) Naming.lookup(resource);
				if (server.acquire(RmiClientParallelizer.this)) {
					Util.pln(name + " -- connected");
					setState(State.READY);
					isAlive = true;
				}
				else {
					Util.pln(name + " -- busy");
					server = null;
				}
			}
			catch (Exception e) {
				Util.pln(name + " -- connection failed: " + e.getMessage());
				e.printStackTrace();
			}
		}
		
		public void close(boolean graceful) {
			if (state.isUp()) {
				try {
					if (graceful)
						server.release();
				}
				catch (RemoteException e) {
					graceful = false;
				}
				finally {
					setState(graceful ? State.CLOSED : State.DIED);
					
					Util.debugln("Closed connection to " + name);
				}
			}
		}
		
		public void remoteClose() {
			// Assumes the server took care of releasing the client itself; the
			// server only closes itself when it encounters a fatal error.
			setState(State.DIED);
			Util.debugln("Closed connection to " + name);
		}
		
		public String getName() {
			return name;
		}
		
		public State getState() {
			return state;
		}
		
		public void setState(State newState) {

			synchronized(RmiClientParallelizer.this) {
				
				if (state == newState)
					return;
				
				if (state.isDown() && newState.isUp())
					numRunning++;
				if (state.isUp() && newState.isDown())
					numRunning--;
				if (state == State.BUSY)
					numBusy--;
				if (newState == State.BUSY)
					numBusy++;
				
				state = newState;
				RmiClientParallelizer.this.notify();
			}
		}
		
		public void setCurrentTask(MovableTask task) {
			setState((task != null) ? State.BUSY : State.READY);
			currentTask = task;
		}
		
		public synchronized void setAlive() {
			isAlive = true;
		}
		
		public synchronized void checkAlive() {
			if (state.isUp() && !isAlive) {
				Util.debugln(name + " hasn't talked to the server recently; assuming it's dead");
				close(false);
			}
			isAlive = false;
		}
		
		public void start() {
			if (state == State.READY) {
				try {
					Util.debugln("Telling " + name + " to start");
					
					server.start(arg, id);
					setState(State.BUSY);
				}
				catch (RemoteException e) {
					handleRemoteException(e);
				}
			}
		}
		
		public MovableTask stealTask() {
			if (state == State.BUSY) {
				try {
					Util.debugln("     Trying to steal from " + name);
					
					MovableTask result = server.stealTask();
					setAlive();
					return result;
				}
				catch (RemoteException e) {
					handleRemoteException(e);
				}
			}
			return null;
		}

		private void handleRemoteException(RemoteException e) {
			Util.pln(name + " -- threw a RemoteException: " + e.getMessage());
			e.printStackTrace();
			
			// Timeouts aren't sufficient cause for assuming the server actually died.
			
			if (e.getCause() == null || !(e.getCause() instanceof SocketTimeoutException)) {
				if (currentTask != null) {
					enqueueTask(currentTask);
					currentTask = null;
				}
				close(false);
			}
		}
	}
	
	//-----------------------------------------------------------------------------------
	//---------------------------LIVENESS CHECKING---------------------------------------
	//-----------------------------------------------------------------------------------
	private class Reaper extends Thread {
		private static final int DELAY = 180000;			// 3 min between sleeps
		
		public Reaper() {
			super("Reaper");
		}
		
		public void run() {
			try {
				do {
					Thread.sleep(DELAY);
					for (int i = 0; i < servers.length; i++)
						servers[i].checkAlive();
				} while (numRunning > 0);
			}
			catch (InterruptedException e) {
				// breaks out of loop
			}
		}
	}
}
