package parallel;

import assign.Solution;
import assign.Worker;
import assign.Worker.Argument;
import util.Util;

import java.net.MalformedURLException;
import java.rmi.Naming;
import java.rmi.RemoteException;
import java.rmi.server.UnicastRemoteObject;


/**
 * Parallelizer that accepts incoming tasks from the RmiClient and
 * allows the client to steal pending tasks.
 * 
 * @author kulinipm
 */
public class RmiServerParallelizer extends Parallelizer implements RmiServer, Runnable {
	
	private static final int FETCH_DELAY = 5000;	// milliseconds; arbitrarily chosen
	
	private RmiClient client;
	private int numThreads;
	
	private Argument clientArg;
	private MovableTaskRunnerGroup group;
	private Thread mainThread;
	private Pinger pinger;
	private int id;
		
	//-----------------------------------------------------------------------------------
	//-----------------------------CONSTRUCTORS------------------------------------------
	//-----------------------------------------------------------------------------------
	
	public RmiServerParallelizer(Argument arg) throws RemoteException, MalformedURLException {
		
		//-----------set up what we can; the protein's information will come later------
		client = null;
		numThreads = arg.threads;
		
		clientArg = null;
		group = null;
		mainThread = null;
		pinger = null;
		id = -1;
		
		//------------------make this server visible via RMI----------------------------

		UnicastRemoteObject.exportObject(this);
		Naming.rebind("//localhost:" + arg.port + "/RmiServer", this);
		
		Util.pln("\n\nWaiting for an incoming connection\n");
	}
	
	//-----------------------------------------------------------------------------------
	//--------------------------SUPERCLASS METHODS---------------------------------------
	//-----------------------------------------------------------------------------------

	public void start() {
		// do nothing; tasks come from the network, not the local machine
	}
	
	public void enqueueTask(MovableTask task) {
		try {
			client.enqueueTask(id, task);
		}
		catch (RemoteException e) {
			handleRemoteException(e);
		}
	}
	
	public MovableTask dequeueTask() {
		try {
			return client.fetchTask(id);
		}
		catch (RemoteException e) {
			handleRemoteException(e);
			return null;
		}
	}
	
	public void abort() {
		// Give the client whatever's been found so far and close the connection.
		try {
			client.addStrands(id, Worker.mappingsFound.getStrands());
			client.close(id);
		}
		catch (RemoteException e) {
			Util.pln("Got a RemoteException: " + e.getMessage());
			e.printStackTrace();
		}
		finally {
			// Relies on the fact that the main thread must be running if this
			// method gets executed.
			mainThread.interrupt();
		}
	}
	
	private void handleRemoteException(RemoteException e) {
		Util.pln("Got a RemoteException: " + e.getMessage());
		e.printStackTrace();
		abort();
	}

	//-----------------------------------------------------------------------------------
	//-------------------------RESOURCE ACQUISITION--------------------------------------
	//-----------------------------------------------------------------------------------

	// Because of all the static data used in the various classes, only one client at a
	// time can use the server, thus this advisory locking mechanism.

	public synchronized boolean acquire(RmiClient client) {
		if (this.client == null) {
			Util.pln("Connected\n");
			this.client = client;
			pinger = new Pinger();
			pinger.start();
			return true;
		}
		else
			return false;
	}

	public synchronized void release() {
		if (mainThread != null)
			mainThread.interrupt();
		if (pinger != null)
			pinger.interrupt();

		if (client != null)
			Util.pln("\n\nConnection closed; waiting for an incoming connection\n");
		
		mainThread = null;
		pinger = null;
		client = null;
		clientArg = null;
		id = -1;
	}
	
	//-----------------------------------------------------------------------------------
	//---------------------------SERVER INTERFACE----------------------------------------
	//-----------------------------------------------------------------------------------

	public void start(Argument arg, int id) {
		clientArg = arg;
		this.id = id;
		mainThread = new Thread(this);
		mainThread.start();		
	}
	
	public MovableTask stealTask() {
		if (group == null)
			return null;
		return group.steal();
	}

	public void quit() {
		Util.pln("Got a request from the client to quit.");
		mainThread.interrupt();
	}
	
	//-----------------------------------------------------------------------------------
	//-----------------------MAIN COMPUTATION THREAD-------------------------------------
	//-----------------------------------------------------------------------------------
	
	public void run() {
		try {
			//--------------Precomputation-------------------
			Worker.initialize(clientArg);
			Worker.computeReferenceSolution();
			Solution.initialize();
		
			//--------------Process tasks--------------------
			group = new MovableTaskRunnerGroup(numThreads);
			while (true) {
				MovableTask task = dequeueTask();
				if (task != null) {
					group.invoke(task);
					client.addStrands(id, Worker.mappingsFound.getStrands());
				}
				else
					Thread.sleep(FETCH_DELAY);
			}
		}
		catch (InterruptedException e) {
			// breaks out of the while loop
		}
		catch (RemoteException e) {
			Util.pln("Got exception from client: " + e.getMessage());
			e.printStackTrace();
		}
		finally {
			if (group != null) {
				group.interruptAll();
				group.stats();
			}
			release();
		}
	}
	
	//-----------------------------------------------------------------------------------
	//----------------------------PINGER THREAD------------------------------------------
	//-----------------------------------------------------------------------------------
	
	private class Pinger extends Thread {
		private static final int DELAY = 60000;				// 1 min between pings
		
		public Pinger() {
			super("Pinger");
			
			// The pinger spends most of its time sleeping, but it's important that it
			// runs when it needs to.
			
			setPriority(NORM_PRIORITY + 1);
		}
		
		public void run() {
			try {
				while (client != null) {
					Thread.sleep(DELAY);
					client.ping(id);
				}
			}
			catch (InterruptedException e) {
				// break out of loop
			}
			catch (RemoteException e) {
				handleRemoteException(e);
			}
		}
	}
}
