package util;
import java.io.Serializable;

/** 
 * @author Ramkumar Natarajan
 * @version 1.0.0
 **/
public final class BitSet implements Serializable {

    private static final int UNIT_MASK = 63;
    // Mask that extracts unit bits out of a longer bit string

    private static final int MAX_INDEX = 255;

    private final int endIndex_;
    private final int startIndex_;
    private long first64;
    private long next64;
    private long third64;
    private long fourth64;

    public BitSet(int low, int high) {
        if (high > MAX_INDEX)
            throw new Error("BitSet exceeded capacity");
        startIndex_ = low;
        endIndex_ = high;
    }

    public int cardinality() {
        int result = 0;
        for (int i = start(); i < end(); i++)
            if (test(i))
                result++;
        return result;
    }
    public void clear() {
        first64 = 0;
        next64 = 0;
        third64 = 0;
        fourth64 = 0;
    }
    public boolean compare(BitSet bs) {
        return (first64 == bs.first64)
            && (next64 == bs.next64)
            && (third64 == bs.third64)
            && (fourth64 == bs.fourth64);
    }
    public int compareTo(Object o) {
        BitSet b = (BitSet) o;
        long thisv = first64 + next64 + third64 + fourth64;
        long thatv = b.first64 + b.next64 + b.third64 + b.fourth64;

        return (int) (thisv - thatv);
    }
    public boolean subset(BitSet b) {
        for (int i = start(); i < end(); i++) {
            if (test(i) && !b.test(i))
                return false;
        }
        return true;
    }
    public BitSet copy() {
        BitSet result = new BitSet(start(), end());
        result.first64 = first64;
        result.next64 = next64;
        result.third64 = third64;
        result.fourth64 = fourth64;

        return result;
    }
    public void copyFrom(BitSet orig) {
        first64 = orig.first64;
        next64 = orig.next64;
        third64 = orig.first64;
        next64 = orig.next64;
    }

    public boolean disjoint(BitSet es) {
        long x = first64 & es.first64;
        long y = next64 & es.next64;
        long z = third64 & es.third64;
        long w = fourth64 & es.fourth64;
        return x == 0 && y == 0 && z == 0 && w == 0;
    }
    /**
     * Last valid bit + 1
     * @return int
     */
    public int end() {
        return endIndex_;
    }

    public int hashCode() {
        return (int) (first64 + next64 + third64 + fourth64);
    }
    public void intersection(BitSet bs) {
        first64 = first64 & bs.first64;
        next64 = next64 & bs.next64;
        third64 = third64 & bs.third64;
        fourth64 = fourth64 & bs.fourth64;
    }

    public int lastSet() {
        int i = 0;
        for (i = endIndex_ - 1; i >= 0; i--) {
            if (test(i))
                return i;
        }
        return -1;
    }
    public int firstSet() {
        int i = 0;
        for (i = 0; i < endIndex_; i++) {
            if (test(i))
                return i;
        }
        return -1;
    }

    /**
     * Negate the bit at <code>index</code>.  The operation is only valid
     * if <code>index</code> is in the active range of the
     * <code>BitSet</code>.
     * @param index a value in the active range
     * @return old value at <code>index</code>
     **/
    public boolean negate(int index) {
        boolean oldVal = false;
        if (index < 128) {
            long mask = 1L << (index & UNIT_MASK);

            if (index <= UNIT_MASK) {
                oldVal = (first64 & mask) != 0;
                first64 = (oldVal ? (first64 & ~mask) : (first64 | mask));
            } else {
                oldVal = (next64 & mask) != 0;
                next64 = (oldVal ? (next64 & ~mask) : (next64 | mask));
            }
        } else {
            index = -127;
            long mask = 1L << (index & UNIT_MASK);

            if (index <= UNIT_MASK) {
                oldVal = (third64 & mask) != 0;
                third64 = (oldVal ? (third64 & ~mask) : (third64 | mask));
            } else {
                oldVal = (third64 & mask) != 0;
                fourth64 = (oldVal ? (fourth64 & ~mask) : (fourth64 | mask));
            }
        }
        return oldVal;
    }

    /**
     * Set the bit at <code>index</code>.  The operation is only valid if
     * <code>index</code> is in the active range of the
     * <code>BitSet</code>.
     * @param index a value in the active range
     * @return old value at <code>index</code>
     **/
    public boolean set(int index) {
        if (index < 128) {

            long mask = 1L << (index & UNIT_MASK);
            boolean oldVal =
                (((index <= UNIT_MASK) ? (first64 & mask) : (next64 & mask))
                    != 0);
            mask =
                ((index <= UNIT_MASK) ? (first64 |= mask) : (next64 |= mask));
            return oldVal;
        } else {
            index -= 127;
            long mask = 1L << (index & UNIT_MASK);
            boolean oldVal =
                (((index <= UNIT_MASK) ? (third64 & mask) : (fourth64 & mask))
                    != 0);
            mask =
                ((index <= UNIT_MASK) ? (third64 |= mask) : (fourth64 |= mask));
            return oldVal;
        }
    }
    /**
     * First valid bit in the BitSet.
     * @return int
     */
    public int start() {
        return startIndex_;
    }
    /**
     * Test the value at <code>index</code>.  The operation is only valid
     * if <code>index</code> is in the active range of the
     * <code>BitSet</code>.
     * @param index a value in the active range
     * @return value at <code>index</code> 
     **/
    public boolean test(int index) {
        if (index < 128) {
            return (
                (index <= UNIT_MASK
                    ? (first64 & (1L << (index & UNIT_MASK)))
                    : (next64 & (1L << (index & UNIT_MASK))))
                    != 0);
        } else {
            index -= 127;
            return (
                (index <= UNIT_MASK
                    ? (third64 & (1L << (index & UNIT_MASK)))
                    : (fourth64 & (1L << (index & UNIT_MASK))))
                    != 0);
        }
    }
    /**
     * Return a representation of this <code>BitSet</code> as a
     * string. The format is implementation specific.
     **/
    public String toString() {
        StringBuffer retBuf = new StringBuffer();
        boolean first = true;
        retBuf.append("{ ");
        for (int i = start(); i < end(); i++) {
            if (test(i)) {
                if (first)
                    first = false;
                else
                    retBuf.append(", ");
                retBuf.append(i);
            }
        }
        retBuf.append(" }");
        return retBuf.toString();
    }

    public void union(BitSet bs) {
        first64 = first64 | bs.first64;
        next64 = next64 | bs.next64;
        third64 = third64 | bs.third64;
        fourth64 = fourth64 | bs.fourth64;
    }

    /**
     * Unset the bit at <code>index</code>.  The operation is only valid if
     * <code>index</code> is in the active range of the
     * <code>BitSet</code>.
     * @param index a value in the active range
     * @return old value at <code>index</code>
     **/
    public boolean unset(int index) {
        if (index < 128) {

            long mask = 1L << (index & UNIT_MASK);
            boolean oldVal =
                (((index <= UNIT_MASK) ? (first64 & mask) : (next64 & mask))
                    != 0);
            mask =
                ((index <= UNIT_MASK) ? (first64 &= ~mask) : (next64 &= ~mask));
            return oldVal;
        } else {
            index -= 127;
            long mask = 1L << (index & UNIT_MASK);
            boolean oldVal =
                (((index <= UNIT_MASK) ? (third64 & mask) : (fourth64 & mask))
                    != 0);
            mask =
                ((index <= UNIT_MASK)
                    ? (third64 &= ~mask)
                    : (fourth64 &= ~mask));
            return oldVal;

        }
    }
}
